from typing import (
    IO,
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
)

import datetime
import distutils.util
import logging
import os
import re
import shlex
import sys
import time
from argparse import (
    SUPPRESS,
    Action,
    ArgumentDefaultsHelpFormatter,
    ArgumentParser,
    FileType,
    HelpFormatter,
    Namespace,
)

import numpy as np

from smac.utils.constants import MAXINT, N_TREES
from smac.utils.io.input_reader import (
    INSTANCE_FEATURES_TYPE,
    INSTANCE_TYPE,
    InputReader,
)

__author__ = "Marius Lindauer"
__copyright__ = "Copyright 2018, ML4AAD"
__license__ = "3-clause BSD"


in_reader = InputReader()
PARSED_SCENARIO_ARGS_TYPE = Dict[
    str,
    Union[str, int, Dict, INSTANCE_TYPE, INSTANCE_FEATURES_TYPE, np.ndarray, List[str]],
]
parsed_scen_args = {}  # type: PARSED_SCENARIO_ARGS_TYPE
# Placeholder logger that will not be used in practice, but which will be replaced by
# the logger onces the parsing logic is instantiated
logger = logging.getLogger(__name__)


def truthy(x: Any) -> bool:
    """Convert x into its truth value."""
    if isinstance(x, bool):
        return x
    elif isinstance(x, (int, float)):
        return x != 0
    elif isinstance(x, str):
        return bool(distutils.util.strtobool(x))
    else:
        return False


def multi_objectives(x: Union[str, List[str]]) -> List[str]:
    """Convert objectives into an array."""
    if isinstance(x, str):
        # Convert a (comma-separated) string into list of strings.
        x = x.replace(", ", ",")
        return x.split(",")
    elif isinstance(x, List):
        return x
    else:
        raise RuntimeError("Expected string or list of strings, got %s" % type(x))


def cost_for_crash(x: Union[str, List]) -> Union[int, float, List[Union[int, float]]]:
    """Convert cost for crash into an array."""
    if isinstance(x, List):
        x = [float(i) for i in x]
    else:
        if isinstance(x, str):
            if x[0] == "[" and x[-1] == "]":
                x = x[1:-1]

            # Convert a (comma-separated) string into list of strings.
            x = x.replace(", ", ",").split(",")

        x = [float(i) for i in x]

    if len(x) == 1:
        return x[0]

    return x


class CheckScenarioFileAction(Action):
    """Check scenario file given by user."""

    def __call__(  # type: ignore[override] # noqa F821
        self,
        parser: ArgumentParser,
        namespace: Namespace,
        values: str,
        option_string: Optional[str] = None,
    ) -> None:
        """Check scenario file given by user."""
        fn = values
        if fn:
            if not os.path.isfile(fn):
                parser.exit(1, "Could not find scenario file: {}".format(fn))
        setattr(namespace, self.dest, values)


class ParseRandomConfigurationChooserAction(Action):
    """Parse random configuration chooser given by user."""

    def __call__(  # type: ignore[override] # noqa F821
        self,
        parser: ArgumentParser,
        namespace: Namespace,
        values: IO,
        option_string: Optional[str] = None,
    ) -> None:
        """Parse random configuration chooser given by user."""
        module_file = values
        module_path = module_file.name
        module_file.close()
        import importlib.util

        spec = importlib.util.spec_from_file_location("smac.custom.random_configuration_chooser", module_path)
        assert spec is not None
        assert spec.loader is not None  # please mypy
        rcc_module = importlib.util.module_from_spec(spec)  # type: ignore
        spec.loader.exec_module(rcc_module)  # type: ignore[attr-defined] # noqa F821
        setattr(namespace, self.dest, rcc_module.RandomConfigurationChooserImpl())  # type: ignore[attr-defined] # noqa F821


class ProcessRunObjectiveAction(Action):
    """Process run objective given by user."""

    def __call__(  # type: ignore[override] # noqa F821
        self,
        parser: ArgumentParser,
        namespace: Namespace,
        values: str,
        option_string: Optional[str] = None,
    ) -> None:
        """Process run objective given by user."""
        if values == "runtime":
            parsed_scen_args["cutoff_time_required"] = {
                "error": '--cutoff-time is required when --run-objective is set to "runtime"'
            }
        setattr(namespace, self.dest, values)


class ParseOverallObjectiveAction(Action):
    """Parse overall objective given by user."""

    def __call__(  # type: ignore[override] # noqa F821
        self,
        parser: ArgumentParser,
        namespace: Namespace,
        values: str,
        option_string: Optional[str] = None,
    ) -> None:
        """Parse overall objective given by user."""
        par_str = values
        if par_str[:3] in ["PAR", "par"]:
            par_str = par_str[3:]
        elif par_str[:4] in ["mean", "MEAN"]:
            par_str = par_str[4:]
        # Check for par-value as in "par10"/ "mean5"
        if par_str:
            parsed_scen_args["par_factor"] = int(par_str)
        else:
            logger.debug("No par-factor detected. Using 1 by default.")
            parsed_scen_args["par_factor"] = 1
        setattr(namespace, self.dest, values)


class ReadTrainInstFileAction(Action):
    """Read training instance file given by user."""

    def __call__(  # type: ignore[override] # noqa F821
        self,
        parser: ArgumentParser,
        namespace: Namespace,
        values: Optional[str],
        option_string: Optional[str] = None,
    ) -> None:
        """Read training instance file given by user."""
        fn = values
        if fn:
            if os.path.isfile(fn):
                parsed_scen_args["train_insts"] = in_reader.read_instance_file(fn)
            else:
                parser.exit(1, "Could not find instance file: {}".format(fn))
        setattr(namespace, self.dest, values)


class ReadTestInstFileAction(Action):
    """Read test instance file given by user."""

    def __call__(  # type: ignore[override] # noqa F821
        self,
        parser: ArgumentParser,
        namespace: Namespace,
        values: Optional[str],
        option_string: Optional[str] = None,
    ) -> None:
        """Read test instance file given by user."""
        fn = values
        if fn:
            if os.path.isfile(fn):
                parsed_scen_args["test_insts"] = in_reader.read_instance_file(fn)
            else:
                parser.exit(1, "Could not find test instance file: {}".format(fn))
        setattr(namespace, self.dest, values)


class ReadFeatureFileAction(Action):
    """Read feature file given by user."""

    def __call__(  # type: ignore[override] # noqa F821
        self,
        parser: ArgumentParser,
        namespace: Namespace,
        values: Optional[str],
        option_string: Optional[str] = None,
    ) -> None:
        """Read feature file given by user."""
        fn = values
        if fn:
            if os.path.isfile(fn):
                instance_features = in_reader.read_instance_features_file(fn)
                (
                    parsed_scen_args["feature_names"],
                    parsed_scen_args["feature_dict"],
                ) = instance_features
                parsed_scen_args["features"] = instance_features
            else:
                parser.exit(1, "Could not find feature file: {}".format(fn))
        setattr(namespace, self.dest, values)


class ReadPCSFileAction(Action):
    """Read PCS (parameter configuration space) file given by user."""

    def __call__(  # type: ignore[override] # noqa F821
        self,
        parser: ArgumentParser,
        namespace: Namespace,
        values: Optional[str],
        option_string: Optional[str] = None,
    ) -> None:
        """Read PCS (parameter configuration space) file given by user."""
        fn = values
        if fn:
            if os.path.isfile(fn):
                cs = in_reader.read_pcs_file(fn)
                cs.seed(42)
                parsed_scen_args["cs"] = cs
            else:
                parser.exit(1, "Could not find pcs file: {}".format(fn))
        setattr(namespace, self.dest, values)


class ProcessOutputDirAction(Action):
    """Process output directory given by user."""

    def __call__(  # type: ignore[override] # noqa F821
        self,
        parser: ArgumentParser,
        namespace: Namespace,
        values: Optional[str],
        option_string: str = None,
    ) -> None:
        """Process output directory given by user."""
        directory = values
        if not directory:
            logger.debug("Deactivate output directory.")
            values = None
        else:
            logger.info("Output to {}".format(directory))
        setattr(namespace, self.dest, values)


class ConfigurableHelpFormatter(ArgumentDefaultsHelpFormatter):
    """Configurable Help Formatter. Can filter out developer options."""

    def __init__(self, *args: Any, help_type: str = "standard", **kwargs: Any):
        self.help_type = help_type
        super(ConfigurableHelpFormatter, self).__init__(*args, **kwargs)

    def _add_item(self, func: Callable, args: Any) -> None:
        """
        Add item to help formatter.

        Parameters
        ----------
        func : Callable
            Function to call.
        args : Any
            Arguments to pass to function.
        """

        def filter_actions(actions: List[Action]) -> List[Action]:
            filtered_actions = []
            for action in actions:
                dev = False
                if isinstance(action.help, str):
                    if action.help.startswith("[dev]"):
                        dev = True
                else:
                    for s in action.option_strings:
                        if s.startswith("--dev"):
                            dev = True
                            break
                if not dev:
                    filtered_actions.append(action)
            return filtered_actions

        if self.help_type == "standard":
            if func.__name__ == "_format_usage":
                args = (args[0], filter_actions(args[1]), args[2], args[3])
            elif isinstance(args, list):
                if args:
                    args = filter_actions(args)
                    if not args:
                        return
        self._current_section.items.append((func, args))


class SMACArgumentParser(ArgumentParser):
    """ArgumentParser that can be extended by additional parsers."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self.additional_parsers = []  # type: List[ArgumentParser]
        self.help_type = "standard"  # standard or dev
        super(SMACArgumentParser, self).__init__(*args, **kwargs)

    def set_help_type(self, help_type: str) -> None:
        """Set help type to `self.additional_parsers`."""
        self.help_type = help_type
        for parser in self.additional_parsers:
            parser.help_type = help_type  # type: ignore[attr-defined] # noqa F821

    def add_parser(self, additional_parser: ArgumentParser) -> None:
        """Add parser to `self.additional_parsers`."""
        self.additional_parsers.append(additional_parser)

    def _get_formatter(self) -> HelpFormatter:
        """Returns formatter from `self.formatter_class`."""
        return self.formatter_class(prog=self.prog, help_type=self.help_type)  # type: ignore[call-arg] # noqa F821

    def format_help(self) -> str:
        """Get help string from formatter."""
        formatter = self._get_formatter()

        # usage
        formatter.add_usage(
            self.usage,  # type: ignore[arg-type] # noqa F821
            self._actions,
            self._mutually_exclusive_groups,
        )

        # description
        formatter.add_text(self.description)

        # positionals, optionals and user-defined groups
        def add_action_groups(parser: ArgumentParser) -> None:
            for action_group in parser._action_groups:
                formatter.start_section(action_group.title)
                formatter.add_text(action_group.description)
                formatter.add_arguments(action_group._group_actions)
                formatter.end_section()

        add_action_groups(self)

        # positionals, optionals and user-defined groups from additional parsers
        for parser in self.additional_parsers:
            add_action_groups(parser)

        # epilog
        formatter.add_text(self.epilog)

        # determine help from format above
        return formatter.format_help()


class StandardHelpAction(Action):
    """Action to only show standard options in help message."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        # https://github.com/python/mypy/issues/6799
        super().__init__(default=SUPPRESS, nargs=0, *args, **kwargs)  # type: ignore

    def __call__(  # type: ignore[override] # noqa F821
        self,
        parser: SMACArgumentParser,
        namespace: Namespace,
        values: list,
        option_string: Optional[str] = None,
    ) -> None:
        """Action to only show standard options in help message."""
        parser.set_help_type("standard")
        parser.print_help()
        parser.exit()


class DevHelpAction(Action):
    """Action to show standard and developer options in help message."""

    def __init__(self, *args: Any, **kwargs: Any):
        # https://github.com/python/mypy/issues/6799
        super().__init__(default=SUPPRESS, nargs=0, *args, **kwargs)  # type: ignore

    def __call__(  # type: ignore[override] # noqa F821
        self,
        parser: SMACArgumentParser,
        namespace: Namespace,
        values: list,
        option_string: Optional[str] = None,
    ) -> None:
        """Action to show standard and developer options in help message."""
        parser.set_help_type("dev")
        parser.print_help()
        parser.exit()


class CMDReader(object):
    """Use argparse to parse command line options.

    Attributes
    ----------
    logger : Logger
    """

    def __init__(self) -> None:
        global logger, parsed_scen_args
        self.logger = logging.getLogger(self.__module__ + "." + self.__class__.__name__)
        logger = self.logger

        # initialized in _add_main_options
        self.parser = None  # type: SMACArgumentParser # type: ignore[assignment]
        self.main_cmd_actions = {}  # type: Dict[str, Dict]
        self.main_cmd_translations = {}  # type: Dict[str, str]
        # initialized in _add_smac_options
        self.smac_parser = None  # type: SMACArgumentParser # type: ignore[assignment]
        self.smac_cmd_actions = {}  # type: Dict[str, Dict]
        self.smac_cmd_translations = {}  # type: Dict[str, str]
        # initialized in _add_scen_options
        self.scen_parser = None  # type: SMACArgumentParser # type: ignore[assignment]
        self.scen_cmd_actions = {}  # type: Dict[str, Dict]
        self.scen_cmd_translations = {}  # type: Dict[str, str]
        # needed for argument interdependencies
        self.parsed_scen_args = {}  # type: PARSED_SCENARIO_ARGS_TYPE
        parsed_scen_args = self.parsed_scen_args

        # add arguments to parser
        self._add_main_options()
        self._add_smac_options()
        self._add_scen_options()

    @staticmethod
    def _extract_action_info(actions: List[Action]) -> Tuple[Dict, Dict]:
        """Extracts action information."""
        extracted_info = {}
        translations = {}
        for action in actions:
            name_list = list(filter(lambda e: e.startswith("--"), action.option_strings))
            if name_list:
                name = name_list[0]
            else:
                name = action.option_strings[0]
            dest = name
            if hasattr(action, "dest"):
                dest = action.dest
            cmd_action = (
                dict()
            )  # type: Dict[str, Union[str, Callable[[str], Any], IO, None, Type, bool, Sequence[str], Iterable]] # noqa 501
            cmd_action["dest"] = dest
            for name in action.option_strings:
                translations[name] = dest
                translations[name.lstrip("-")] = dest
            if hasattr(action, "type"):
                cmd_action["type"] = action.type
            else:
                cmd_action["type"] = str
            if hasattr(action, "default"):
                if action.default == SUPPRESS:
                    continue
                cmd_action["default"] = action.default
            else:
                cmd_action["default"] = None
            if hasattr(action, "choices"):
                cmd_action["choices"] = action.choices
            else:
                cmd_action["choices"] = None
            if hasattr(action, "required"):
                cmd_action["required"] = action.required
            else:
                cmd_action["required"] = False
            if hasattr(action, "help"):
                cmd_action["help"] = action.help
            else:
                cmd_action["help"] = None
            cmd_action["option_strings"] = action.option_strings
            extracted_info[name] = cmd_action
        return extracted_info, translations

    def _add_main_options(self) -> None:
        """Add main options."""
        prog = sys.argv[0]
        if re.match("^python[0-9._-]*$", sys.argv[0]):
            prog = sys.argv[1]
        self.parser = SMACArgumentParser(formatter_class=ConfigurableHelpFormatter, add_help=False, prog=prog)
        # let a help message begin with "[dev]" to add a developer option
        req_opts = self.parser.add_argument_group("Required Options")
        req_opts.add_argument(
            "--scenario",
            "--scenario-file",
            "--scenario_file",
            dest="scenario_file",
            required=True,
            type=str,
            action=CheckScenarioFileAction,
            help="Scenario file in AClib format.",
        )
        opt_opts = self.parser.add_argument_group("Optional Options")
        opt_opts.add_argument(
            "--help",
            action=StandardHelpAction,
            help="Show help messages for standard options.",
        )
        opt_opts.add_argument(
            "--help-all",
            action=DevHelpAction,
            help="Show help messages for both standard and developer options.",
        )
        opt_opts.add_argument("--seed", default=1, type=int, help="Random Seed.")
        opt_opts.add_argument(
            "--verbose",
            "--verbose-level",
            "--verbose_level",
            dest="verbose_level",
            default=logging.INFO,
            choices=["INFO", "DEBUG"],
            help="Verbosity level.",
        )
        opt_opts.add_argument(
            "--mode",
            default="SMAC4AC",
            choices=["SMAC4AC", "ROAR", "Hydra", "PSMAC", "SMAC4HPO", "SMAC4BB"],
            help="Configuration mode.",
        )
        opt_opts.add_argument(
            "--restore-state",
            "--restore_state",
            dest="restore_state",
            default=None,
            help="Path to directory with SMAC-files.",
        )
        # list of runhistory dump files
        # scenario corresponding to --warmstart_runhistory;
        # pcs and feature space has to be identical to --scenario_file
        opt_opts.add_argument(
            "--warmstart-runhistory",
            "--warmstart_runhistory",
            dest="warmstart_runhistory",
            default=None,
            nargs="*",
            help=SUPPRESS,
        )
        opt_opts.add_argument(
            "--warmstart-scenario",
            "--warmstart_scenario",
            dest="warmstart_scenario",
            default=None,
            nargs="*",
            help=SUPPRESS,
        )
        # list of trajectory dump files, reads runhistory and uses final incumbent as challenger
        opt_opts.add_argument(
            "--warmstart-incumbent",
            "--warmstart_incumbent",
            dest="warmstart_incumbent",
            default=None,
            nargs="*",
            help=SUPPRESS,
        )
        req_opts.add_argument(
            "--random_configuration_chooser",
            default=None,
            type=FileType("r"),
            help="[dev] path to a python module containing a class `RandomConfigurationChooserImpl`"
            "implementing the interface of `RandomConfigurationChooser`",
        )
        req_opts.add_argument(
            "--hydra_iterations",
            default=3,
            type=int,
            help="[dev] number of hydra iterations. Only active if mode is set to Hydra",
        )
        req_opts.add_argument(
            "--hydra_validation",
            default="train",
            choices=["train", "val10", "val20", "val30", "val40", "val50", "none"],
            type=str.lower,
            help="[dev] set to validate incumbents on. valX =>" " validation set of size training_set * 0.X",
        )
        req_opts.add_argument(
            "--incumbents_per_round",
            default=1,
            type=int,
            help="[dev] number of configurations to keep per psmac/hydra iteration.",
            dest="hydra_incumbents_per_round",
        )
        req_opts.add_argument(
            "--n_optimizers",
            default=1,
            type=int,
            help="[dev] number of optimizers to run in parallel per psmac/hydra iteration.",
            dest="hydra_n_optimizers",
        )
        req_opts.add_argument(
            "--psmac_validate",
            default=False,
            type=truthy,
            help="[dev] Validate all psmac configurations.",
        )

        (
            self.main_cmd_actions,
            self.main_cmd_translations,
        ) = CMDReader._extract_action_info(self.parser._actions)

    def _add_smac_options(self) -> None:
        """Add SMAC options."""
        self.smac_parser = SMACArgumentParser(formatter_class=ConfigurableHelpFormatter, add_help=False)
        smac_opts = self.smac_parser.add_argument_group("SMAC Options")
        smac_opts.add_argument(
            "--abort-on-first-run-crash",
            "--abort_on_first_run_crash",
            dest="abort_on_first_run_crash",
            default=True,
            type=truthy,
            help="If true, *SMAC* will abort if the first run of " "the target algorithm crashes.",
        )
        smac_opts.add_argument(
            "--limit-resources",
            "--limit_resources",
            dest="limit_resources",
            default=False,
            type=truthy,
            help="If true, *SMAC* will use pynisher to limit time and memory for "
            "the target algorithm. Allows SMAC to use all resources available. "
            "Applicable only to func TAEs. Set to 'False' by default. "
            "(Warning: This only works on Linux. Use with caution!)",
        )
        smac_opts.add_argument(
            "--minr",
            "--minR",
            dest="minR",
            default=1,
            type=int,
            help="[dev] Minimum number of calls per configuration.",
        )
        smac_opts.add_argument(
            "--maxr",
            "--maxR",
            dest="maxR",
            default=2000,
            type=int,
            help="[dev] Maximum number of calls per configuration.",
        )
        self.output_dir_arg = smac_opts.add_argument(
            "--output-dir",
            "--output_dir",
            dest="output_dir",
            type=str,
            action=ProcessOutputDirAction,
            default="smac3-output_%s" % (datetime.datetime.fromtimestamp(time.time()).strftime("%Y-%m-%d_%H:%M:%S_%f")),
            help="Specifies the output-directory for all emerging " "files, such as logging and results.",
        )
        smac_opts.add_argument(
            "--input-psmac-dirs",
            "--input_psmac_dirs",
            dest="input_psmac_dirs",
            default=None,
            help="For parallel SMAC, multiple output-directories " "are used.",
        )  # TODO: type (list of strings? --> str, nargs=*)
        smac_opts.add_argument(
            "--shared-model",
            "--shared_model",
            dest="shared_model",
            default=False,
            type=truthy,
            help="Whether to run SMAC in parallel mode.",
        )
        smac_opts.add_argument(
            "--random-configuration-chooser",
            "--random_configuration_chooser",
            dest="random_configuration_chooser",
            default=None,
            type=FileType("r"),
            action=ParseRandomConfigurationChooserAction,
            help="[dev] path to a python module containing a class"
            "`RandomConfigurationChooserImpl` implementing"
            "the interface of `RandomConfigurationChooser`",
        )
        smac_opts.add_argument(
            "--hydra-iterations",
            "--hydra_iterations",
            dest="hydra_iterations",
            default=3,
            type=int,
            help="[dev] number of hydra iterations. Only active if mode is set to Hydra",
        )
        smac_opts.add_argument(
            "--use-ta-time",
            "--use_ta_time",
            dest="use_ta_time",
            default=False,
            type=truthy,
            help="[dev] Instead of measuring SMAC's wallclock time, "
            "only consider time reported by the target algorithm (ta).",
        )

        # Hyperparameters
        smac_opts.add_argument(
            "--always-race-default",
            "--always_race_default",
            dest="always_race_default",
            default=False,
            type=truthy,
            help="[dev] Race new incumbents always against default " "configuration.",
        )
        smac_opts.add_argument(
            "--intensification-percentage",
            "--intensification_percentage",
            dest="intensification_percentage",
            default=0.5,
            type=float,
            help="[dev] The fraction of time to be used on "
            "intensification (versus choice of next "
            "Configurations).",
        )
        smac_opts.add_argument(
            "--transform_y",
            "--transform-y",
            dest="transform_y",
            choices=["NONE", "LOG", "LOGS", "INVS"],
            default="NONE",
            help="[dev] Transform all observed cost values"
            " via log-transformations or inverse scaling."
            ' The subfix "s" indicates that SMAC scales the'
            " y-values accordingly to apply the transformation.",
        )

        # RF Hyperparameters
        smac_opts.add_argument(
            "--rf_num_trees",
            "--rf-num-trees",
            dest="rf_num_trees",
            default=N_TREES,
            type=int,
            help="[dev] Number of trees in the random forest (> 1).",
        )
        smac_opts.add_argument(
            "--rf_do_bootstrapping",
            "--rf-do-bootstrapping",
            dest="rf_do_bootstrapping",
            default=True,
            type=bool,
            help="[dev] Use bootstraping in random forest.",
        )
        smac_opts.add_argument(
            "--rf_ratio_features",
            "--rf-ratio-features",
            dest="rf_ratio_features",
            default=5.0 / 6.0,
            type=float,
            help="[dev] Ratio of sampled features in each split ([0.,1.]).",
        )
        smac_opts.add_argument(
            "--rf_min_samples_split",
            "--rf-min-samples-split",
            dest="rf_min_samples_split",
            default=3,
            type=int,
            help="[dev] Minimum number of samples" " to split for building a tree in the random forest.",
        )
        smac_opts.add_argument(
            "--rf_min_samples_leaf",
            "--rf-min-samples-leaf",
            dest="rf_min_samples_leaf",
            default=3,
            type=int,
            help="[dev] Minimum required number of" " samples in each leaf of a tree in the random forest.",
        )
        smac_opts.add_argument(
            "--rf_max_depth",
            "--rf-max-depth",
            dest="rf_max_depth",
            default=20,
            type=int,
            help="[dev] Maximum depth of each tree in the random forest.",
        )
        # AcquisitionOptimizer SLS
        smac_opts.add_argument(
            "--sls_n_steps_plateau_walk",
            "--sls-n-steps-plateau-walk",
            dest="sls_n_steps_plateau_walk",
            default=10,
            type=int,
            help="[dev] Maximum number of steps on plateaus during " "the optimization of the acquisition function.",
        )
        smac_opts.add_argument(
            "--sls_max_steps",
            "--sls-max-steps",
            dest="sls_max_steps",
            default=None,
            type=int,
            help="[dev] Maximum number of local search steps in one iteration"
            " during the optimization of the acquisition function.",
        )
        smac_opts.add_argument(
            "--acq_opt_challengers",
            "--acq-opt-challengers",
            dest="acq_opt_challengers",
            default=5000,
            type=int,
            help="[dev] Number of challengers returned by acquisition function"
            " optimization. Also influences the number of randomly sampled"
            " configurations to optimized the acquisition function",
        )

        # Intensification
        smac_opts.add_argument(
            "--intens_adaptive_capping_slackfactor",
            "--intens-adaptive-capping-slackfactork",
            dest="intens_adaptive_capping_slackfactor",
            default=1.2,
            type=float,
            help="[dev] Slack factor of adpative capping (factor * adpative cutoff)."
            " Only active if obj is runtime."
            " If set to very large number it practically deactivates adaptive capping.",
        )
        smac_opts.add_argument(
            "--intens_min_chall",
            "--intens-min-chall",
            dest="intens_min_chall",
            default=2,
            type=int,
            help="[dev] Minimal number of challengers to be"
            " considered in each intensification run (> 1)."
            " Set to 1 and in combination with very small intensification-percentage."
            " it will deactivate randomly sampled configurations"
            " (and hence, extrapolation of random forest will be an issue.)",
        )
        smac_opts.add_argument(
            "--rand_prob",
            "--rand-prob",
            dest="rand_prob",
            default=0.5,
            type=float,
            help="[dev] probablity to run a random configuration"
            " instead of configuration optimized on the acquisition function",
        )
        self.parser.add_parser(self.smac_parser)
        (
            self.smac_cmd_actions,
            self.smac_cmd_translations,
        ) = CMDReader._extract_action_info(self.smac_parser._actions)

    def _add_scen_options(self) -> None:
        """Add scenario options."""
        self.scen_parser = SMACArgumentParser(formatter_class=ConfigurableHelpFormatter, add_help=False)
        scen_opts = self.scen_parser.add_argument_group("Scenario Options")
        scen_opts.add_argument(
            "--algo",
            "--ta",
            dest="ta",
            type=shlex.split,
            help="[dev] Specifies the target algorithm call that *SMAC* "
            "will optimize. Interpreted as a bash-command.",
        )
        scen_opts.add_argument(
            "--execdir",
            dest="execdir",
            default=".",
            type=str,
            help="[dev] Specifies the path to the execution-directory.",
        )
        scen_opts.add_argument(
            "--deterministic",
            dest="deterministic",
            default=True,
            type=truthy,
            help="[dev] If true, SMAC assumes that the target function or algorithm is deterministic"
            " (the same static seed of 0 is always passed to the function/algorithm)."
            " If false, different random seeds are passed to the target function/algorithm.",
        )
        scen_opts.add_argument(
            "--run-obj",
            "--run_obj",
            dest="run_obj",
            type=str,
            action=ProcessRunObjectiveAction,
            required=True,
            choices=["runtime", "quality"],
            help="[dev] Defines what metric to optimize. When "
            "optimizing runtime, *cutoff_time* is "
            "required as well.",
        )
        scen_opts.add_argument(
            "--multi-objectives",
            "--multi_objectives",
            dest="multi_objectives",
            default="cost",
            type=multi_objectives,
            help="List of string or comma-separated strings of objectives to optimize.",
        )
        self.overall_obj_arg = scen_opts.add_argument(
            "--overall-obj",
            "--overall_obj",
            dest="overall_obj",
            type=str,
            action=ParseOverallObjectiveAction,
            default="par10",
            help="[dev] PARX, where X is an integer defining the "
            "penalty imposed on timeouts (i.e. runtimes that "
            "exceed the *cutoff-time*).",
        )

        scen_opts.add_argument(
            "--save-instantly",
            "--save_instantly",
            "--save-results-instantly",
            dest="save_instantly",
            default=True,
            type=truthy,
            help="If true, runhistory and stats are saved immediately on changes. "
            "Otherwise, runhistory and states are only saved once after the optimization "
            "process has finished.",
        )
        scen_opts.add_argument(
            "--par-factor",
            "--par_factor",
            dest="par_factor",
            type=float,
            default=10.0,
            help=SUPPRESS,
        )  # added after parsing --overall-obj
        scen_opts.add_argument(
            "--cost-for-crash",
            "--cost_for_crash",
            dest="cost_for_crash",
            default=float(MAXINT),
            type=cost_for_crash,
            help="[dev] Defines the cost-value for crashed runs "
            "on scenarios with quality as run-obj. "
            "If multi-objective is used, a list or comma separated string is accepted too.",
        )
        scen_opts.add_argument(
            "--cutoff-time",
            "--cutoff_time",
            "--cutoff",
            dest="cutoff",
            default=None,
            type=float,
            help="[dev] Maximum runtime, after which the "
            "target algorithm is cancelled. **Required "
            "if *run_obj* is runtime.**",
        )
        scen_opts.add_argument(
            "--memory-limit",
            "--memory_limit",
            dest="memory_limit",
            type=float,
            help="[dev] Maximum available memory the target algorithm " "can occupy before being cancelled in MB.",
        )
        scen_opts.add_argument(
            "--tuner-timeout",
            "--tuner_timeout",
            "--algo-runs-timelimit",
            "--algo_runs_timelimit",
            dest="algo_runs_timelimit",
            default=float("inf"),
            type=float,
            help="[dev] Maximum amount of CPU-time used for optimization.",
        )
        scen_opts.add_argument(
            "--wallclock-limit",
            "--wallclock_limit",
            dest="wallclock_limit",
            default=float("inf"),
            type=float,
            help="[dev] Maximum amount of wallclock-time used for optimization.",
        )
        scen_opts.add_argument(
            "--always-race-default",
            "--always_race_default",
            dest="always_race_default",
            default=False,
            type=truthy,
            help="[dev] Race new incumbents always against default configuration.",
        )
        scen_opts.add_argument(
            "--runcount-limit",
            "--runcount_limit",
            "--ta-run-limit",
            "--ta_run_limit",
            dest="ta_run_limit",
            default=float("inf"),
            type=float,
            help="[dev] Maximum number of algorithm-calls during optimization.",
        )
        scen_opts.add_argument(
            "--instance-file",
            "--instance_file",
            "--train-inst-fn",
            "--train_inst_fn",
            dest="train_inst_fn",
            type=str,
            action=ReadTrainInstFileAction,
            help="[dev] Specifies the file with the training-instances.",
        )
        scen_opts.add_argument(
            "--instances",
            "--train-insts",
            "--train_insts",
            dest="train_insts",
            default=[[None]],  # overridden by --instance-file
            help=SUPPRESS,
        )
        scen_opts.add_argument(
            "--test-instance-file",
            "--test_instance_file",
            "--test-inst-fn",
            "--test_inst_fn",
            dest="test_inst_fn",
            type=str,
            action=ReadTestInstFileAction,
            help="[dev] Specifies the file with the test-instances.",
        )
        scen_opts.add_argument(
            "--test-instances",
            "--test_instances",
            "--test-insts",
            "--test_insts",
            dest="test_insts",
            default=[[None]],  # overridden by --test-instance-file
            help=SUPPRESS,
        )
        scen_opts.add_argument(
            "--feature-file",
            "--feature_file",
            "--feature-fn",
            "--feature_fn",
            dest="feature_fn",
            type=str,
            action=ReadFeatureFileAction,
            help="[dev] Specifies the file with the instance-features.",
        )
        scen_opts.add_argument(
            "--features",
            "--feature-dict",
            "--feature_dict",
            dest="feature_dict",
            default={},  # instance name -> feature vector, overridden by --feature-file
            help=SUPPRESS,
        )
        scen_opts.add_argument(
            "--feature-names",
            "--feature_names",
            dest="feature_names",
            type=list,  # type: ignore[arg-type] # noqa F821
            help=SUPPRESS,
        )  # added after parsing --features
        scen_opts.add_argument(
            "--initial-incumbent",
            "--initial_incumbent",
            dest="initial_incumbent",
            default="DEFAULT",
            type=str,
            choices=["DEFAULT", "RANDOM", "LHD", "SOBOL", "FACTORIAL"],
            help="[dev] DEFAULT is the default from the PCS.",
        )
        scen_opts.add_argument(
            "--paramfile",
            "--param-file",
            "--param_file",
            "--pcs-fn",
            "--pcs_fn",
            dest="pcs_fn",
            type=str,
            action=ReadPCSFileAction,
            help="[dev] Specifies the path to the " "PCS-file.",
        )
        scen_opts.add_argument(
            "--cs",
            default=None,  # ConfigSpace object, overridden by --paramfile
            help=SUPPRESS,
        )

        self.parser.add_parser(self.scen_parser)
        (
            self.scen_cmd_actions,
            self.scen_cmd_translations,
        ) = CMDReader._extract_action_info(self.scen_parser._actions)

    def parse_main_command(self, main_cmd_opts: Sequence[str]) -> Tuple[Namespace, List[str]]:
        """Parse main options."""
        args_, misc = self.parser.parse_known_args(main_cmd_opts)
        try:
            misc.remove(self.parser.prog)
        except ValueError:
            pass
        return args_, misc

    def parse_smac_command(
        self,
        smac_dict: dict = {},
        smac_cmd_opts: List[str] = [],
    ) -> Tuple[Namespace, Dict, List[str]]:
        """Parse SMAC options."""
        # transform smac dict to smac_args
        try:
            smac_cmd_opts.remove(self.parser.prog)
        except ValueError:
            pass
        smac_cmd = []
        misc_dict = {}

        parsed_smac_args = {}
        for k, v in smac_dict.items():
            if k in self.smac_cmd_translations:
                if not isinstance(
                    v,
                    (
                        str,
                        bool,
                        int,
                        float,
                    ),
                ):
                    parsed_smac_args[self.smac_cmd_translations[k]] = v
                else:
                    smac_cmd.append("--" + k.replace("_", "-"))
                    smac_cmd.append(v)
            else:
                misc_dict[k] = v
        smac_cmd.extend(smac_cmd_opts)

        args_, misc_cmd = self.smac_parser.parse_known_args([str(e) for e in smac_cmd])

        # execute output_dir action for default value
        if args_.output_dir == self.output_dir_arg.default:
            self.output_dir_arg(self.parser, args_, self.output_dir_arg.default)

        if args_.shared_model and args_.input_psmac_dirs is None:
            # per default, we assume that
            # all pSMAC runs write to the default output dir
            args_.input_psmac_dirs = "smac3-output*/run*/"

        # copy parsed smac args to smac_args_
        for k, v in parsed_smac_args.items():
            setattr(args_, k, v)

        return args_, misc_dict, misc_cmd

    def parse_scenario_command(
        self,
        scenario_file: str = None,
        scenario_dict: dict = {},
        scenario_cmd_opts: List[str] = [],
    ) -> Namespace:
        """Parses scenario options.

        Parameters
        ----------
        scenario_file : str, optional
            Path to the scenario file, by default None
        scenario_dict : dict, optional
            Mappings of scenario options to values, by default {}
        scenario_cmd_opts : List[str], optional
            Scenario options from command line, by default []

        Returns
        -------
        Namespace
            Parsed scenario arguments.

        Raises
        ------
        ValueError
            If scenario is not a string or dictionary.
        """
        # read scenario file
        scenario_file_dict = {}  # type: Dict[str, Any]
        if isinstance(scenario_file, str):
            scenario_file_dict = in_reader.read_scenario_file(scenario_file)
        elif scenario_file is None:
            pass
        else:
            raise ValueError("Scenario has to be a string or a dictionary")
        # add options from scenario dict (with overwriting)
        scenario_file_dict.update(scenario_dict)
        # transform scenario dict to scen_args
        scen_dict = scenario_file_dict
        scen_cmd = []
        misc_dict = {}

        self.parsed_scen_args.clear()
        for k, v in scen_dict.items():
            if k in self.scen_cmd_translations:
                if not isinstance(
                    v,
                    (
                        str,
                        bool,
                        int,
                        float,
                    ),
                ):
                    # e.g. train_insts, test_insts, cs, features
                    self.parsed_scen_args[self.scen_cmd_translations[k]] = v
                else:
                    scen_cmd.append("--" + k.replace("_", "-"))
                    scen_cmd.append(str(v))
            else:
                misc_dict[k] = v
        scen_cmd.extend(scenario_cmd_opts)

        if misc_dict.keys():
            self.logger.warning("Adding unsupported scenario options: {}".format(misc_dict))
            for k, v in misc_dict.items():
                self.parsed_scen_args[k] = v
            # Fail in a later version:
            # self.scen_parser.exit(1, 'Error: Unknown arguments: {}'.format(misc_dict))

        # append rest of arguments (= override) options from scenario file and
        # parse them with the scenario parser
        scen_args_, misc = self.scen_parser.parse_known_args([str(e) for e in scen_cmd])

        if misc:
            self.scen_parser.exit(1, "Error: Can not parse arguments: {}".format(misc))

        # execute overall_obj action for default value
        if scen_args_.overall_obj == self.overall_obj_arg.default:
            self.overall_obj_arg(self.scen_parser, scen_args_, self.overall_obj_arg.default)

        # make checks that argparse can't perform natively

        if "cutoff_time_required" in self.parsed_scen_args:
            if self.parsed_scen_args["cutoff_time_required"] and not scen_args_.cutoff:
                self.parser.exit(1, "Error: {}".format(self.parsed_scen_args["cutoff_time_required"]["error"]))  # type: ignore # noqa 501
            self.parsed_scen_args.pop("cutoff_time_required")

        # copy parsed scenario args to scen_args_
        # par_factor, train_insts, test_insts, (features, features_names), feature_dict, cs
        for k, v in self.parsed_scen_args.items():
            setattr(scen_args_, k, v)

        return scen_args_

    def read_smac_scenario_dict_cmd(
        self,
        dict_cmd: dict,
        scenario_file: str = None,
    ) -> Tuple[Namespace, Namespace]:
        """Reads smac and scenario options provided in a dictionary.

        Returns
        -------
        smac_args_, scen_args_: smac and scenario options parsed with corresponding ArgumentParser
        """
        smac_args_, misc_dict, misc_cmd = self.parse_smac_command(smac_dict=dict_cmd)
        scen_args_ = self.parse_scenario_command(
            scenario_file=scenario_file,
            scenario_dict=misc_dict,
            scenario_cmd_opts=misc_cmd,
        )

        return smac_args_, scen_args_

    def read_cmd(
        self,
        commandline_arguments: Sequence[str] = tuple(sys.argv[1:]),
    ) -> Tuple[Namespace, Namespace, Namespace]:
        """Reads command line options (main, smac and scenario options)

        Returns
        -------
        smac_args_, scen_args_: parsed arguments;
            main, smac and scenario options parsed with corresponding ArgumentParser
        """
        main_args_, misc = self.parse_main_command(main_cmd_opts=commandline_arguments)
        smac_args_, misc_dict, misc_cmd = self.parse_smac_command(smac_cmd_opts=misc)
        scen_args_ = self.parse_scenario_command(
            scenario_file=main_args_.scenario_file,
            scenario_dict=misc_dict,
            scenario_cmd_opts=misc_cmd,
        )
        return main_args_, smac_args_, scen_args_

    @staticmethod
    def _write_options_to_doc(_arguments: dict, path: str, exclude: List[str]) -> None:
        """Writes options to the documentation."""
        with open(path, "w") as fh:
            for arg in sorted(_arguments.keys()):
                print_arg = arg.lstrip("-").replace("-", "_")
                if print_arg in exclude:
                    continue
                if _arguments[arg]["help"] == SUPPRESS:
                    continue
                fh.write(":{}: ".format(print_arg))
                fh.write("{}".format(_arguments[arg]["help"].lstrip("[dev] ")))
                if "default" in _arguments[arg] and _arguments[arg]["default"]:
                    fh.write(" Default: {}.".format(_arguments[arg]["default"]))
                if "choice" in _arguments[arg] and _arguments[arg]["choice"]:
                    fh.write(" Must be from: {}.".format(_arguments[arg]["choice"]))
                fh.write("\n")
            fh.write("\n\n")

    def write_main_options_to_doc(self, path: str = "main_options.rst") -> None:
        """Writes the SMAC option-list to file for autogeneration in documentation.
        The list is created in doc/conf.py and read in doc/options.rst.

        Parameters
        ----------
        path: string
            Where to write to (relative to doc-folder since executed in conf.py)
        """
        exclude = []  # type: List
        _arguments = self.main_cmd_actions
        CMDReader._write_options_to_doc(_arguments, path, exclude)

    def write_smac_options_to_doc(self, path: str = "smac_options.rst") -> None:
        """Writes the SMAC option-list to file for autogeneration in documentation.
        The list is created in doc/conf.py and read in doc/options.rst.

        Parameters
        ----------
        path: string
            Where to write to (relative to doc-folder since executed in conf.py)
        """
        exclude = []  # type: List
        _arguments = self.smac_cmd_actions
        CMDReader._write_options_to_doc(_arguments, path, exclude)

    def write_scenario_options_to_doc(self, path: str = "scenario_options.rst") -> None:
        """Writes the Scenario option-list to file for autogeneration in documentation.
        The list is created in doc/conf.py and read in doc/options.rst.

        Parameters
        ----------
        path: string
            Where to write to (relative to doc-folder since executed in conf.py)
        """
        exclude = ["cs", "features", "instances", "test_instances"]
        _arguments = self.scen_cmd_actions
        CMDReader._write_options_to_doc(_arguments, path, exclude)



import datetime
import os
import sys
import warnings

name = "SMAC3"
package_name = "smac"
author = (
    "\tMarius Lindauer, Katharina Eggensperger, Matthias Feurer, André Biedenkapp, "
    "Difan Deng,\n\tCarolin Benjamins, Tim Ruhkopf, René Sass and Frank Hutter"
)

author_email = "fh@cs.uni-freiburg.de"
description = "SMAC3, a Python implementation of 'Sequential Model-based Algorithm Configuration'."
url = "https://www.automl.org/"
project_urls = {
    "Documentation": "https://https://github.com/automl.github.io/SMAC3/main",
    "Source Code": "https://github.com/https://github.com/automl/smac",
}
copyright = f"""
    Copyright {datetime.date.today().strftime('%Y')}, Marius Lindauer, Katharina Eggensperger,
    Matthias Feurer, André Biedenkapp, Difan Deng, Carolin Benjamins, Tim Ruhkopf, René Sass
    and Frank Hutter
"""
version = "1.3.4"


if os.name != "posix":
    warnings.warn(
        f"Detected unsupported operating system: {sys.platform}."
        "Please be aware, that SMAC might not run on this system."
    )



from typing import Callable, Dict, List, Optional, Tuple, Union, cast

import inspect
import math
import time
import traceback

import numpy as np
import pynisher

from smac.configspace import Configuration
from smac.stats.stats import Stats
from smac.tae import StatusType
from smac.tae.serial_runner import SerialRunner
from smac.utils.constants import MAX_CUTOFF, MAXINT
from smac.utils.logging import PickableLoggerAdapter

__author__ = "Marius Lindauer, Matthias Feurer"
__copyright__ = "Copyright 2015, ML4AAD"
__license__ = "3-clause BSD"
__maintainer__ = "Marius Lindauer"
__email__ = "lindauer@cs.uni-freiburg.de"
__version__ = "0.0.2"


class AbstractTAFunc(SerialRunner):
    """Baseclass to execute target algorithms which are python functions.

    **Note:*** Do not use directly

    Parameters
    ----------
    ta : callable
        Function (target algorithm) to be optimized.
    stats: Stats()
         stats object to collect statistics about runtime and so on
    multi_objectives: List[str]
        names of the objectives, by default it is a single objective parameter "cost"
    run_obj: str
        run objective of SMAC
    memory_limit : int, optional
        Memory limit (in MB) that will be applied to the target algorithm.
    par_factor: int
        penalization factor
    cost_for_crash : float
        cost that is used in case of crashed runs (including runs
        that returned NaN or inf)
    use_pynisher: bool
        use pynisher to limit resources;
        if disabled
          * TA func can use as many resources
          as it wants (time and memory) --- use with caution
          * all runs will be returned as SUCCESS if returned value is not None

    Attributes
    ----------
    memory_limit
    use_pynisher
    """

    def __init__(
        self,
        ta: Callable,
        stats: Stats,
        multi_objectives: List[str] = ["cost"],
        run_obj: str = "quality",
        memory_limit: Optional[int] = None,
        par_factor: int = 1,
        cost_for_crash: float = float(MAXINT),
        abort_on_first_run_crash: bool = False,
        use_pynisher: bool = True,
    ):

        super().__init__(
            ta=ta,
            stats=stats,
            multi_objectives=multi_objectives,
            run_obj=run_obj,
            par_factor=par_factor,
            cost_for_crash=cost_for_crash,
            abort_on_first_run_crash=abort_on_first_run_crash,
        )
        self.ta = ta
        self.stats = stats
        self.multi_objectives = multi_objectives
        self.run_obj = run_obj

        self.par_factor = par_factor
        self.cost_for_crash = cost_for_crash
        self.abort_on_first_run_crash = abort_on_first_run_crash

        signature = inspect.signature(ta).parameters
        self._accepts_seed = "seed" in signature.keys()
        self._accepts_instance = "instance" in signature.keys()
        self._accepts_budget = "budget" in signature.keys()
        if not callable(ta):
            raise TypeError("Argument `ta` must be a callable, but is %s" % type(ta))
        self._ta = cast(Callable, ta)

        if memory_limit is not None:
            memory_limit = int(math.ceil(memory_limit))
        self.memory_limit = memory_limit

        self.use_pynisher = use_pynisher

        self.logger = PickableLoggerAdapter(self.__module__ + "." + self.__class__.__name__)

    def run(
        self,
        config: Configuration,
        instance: Optional[str] = None,
        cutoff: Optional[float] = None,
        seed: int = 12345,
        budget: Optional[float] = None,
        instance_specific: str = "0",
    ) -> Tuple[StatusType, float, float, Dict]:
        """Runs target algorithm <self._ta> with configuration <config> for at most <cutoff>
        seconds, allowing it to use at most <memory_limit> RAM.

        Whether the target algorithm is called with the <instance> and
        <seed> depends on the subclass implementing the actual call to
        the target algorithm.

        Parameters
        ----------
            config : Configuration, dictionary (or similar)
                Dictionary param -> value
            instance : str, optional
                Problem instance
            cutoff : float, optional
                Wallclock time limit of the target algorithm. If no value is
                provided no limit will be enforced. It is casted to integer internally.
            seed : int
                Random seed
            budget : float, optional
                A positive, real-valued number representing an arbitrary limit to the target algorithm
                Handled by the target algorithm internally
            instance_specific: str
                Instance specific information (e.g., domain file or solution)

        Returns
        -------
            status: enum of StatusType (int)
                {SUCCESS, TIMEOUT, CRASHED, ABORT}
            cost: np.ndarray
                cost/regret/quality/runtime (float) (None, if not returned by TA)
            runtime: float
                runtime (None if not returned by TA)
            additional_info: dict
                all further additional run information
        """
        obj_kwargs = {}  # type: Dict[str, Union[int, str, float, None]]
        if self._accepts_seed:
            obj_kwargs["seed"] = seed
        if self._accepts_instance:
            obj_kwargs["instance"] = instance
        if self._accepts_budget:
            obj_kwargs["budget"] = budget

        cost = self.cost_for_crash  # type: Union[float, List[float]]

        if self.use_pynisher:
            # walltime for pynisher has to be a rounded up integer
            if cutoff is not None:
                cutoff = int(math.ceil(cutoff))
                if cutoff > MAX_CUTOFF:
                    raise ValueError(
                        "%d is outside the legal range of [0, 65535] "
                        "for cutoff (when using pynisher, due to OS limitations)" % cutoff
                    )

            arguments = {
                "logger": self.logger,
                "wall_time_in_s": cutoff,
                "mem_in_mb": self.memory_limit,
            }

            # call ta
            try:
                obj = pynisher.enforce_limits(**arguments)(self._ta)
                rval = self._call_ta(obj, config, obj_kwargs)
            except Exception as e:
                cost = np.asarray(cost).squeeze().tolist()
                exception_traceback = traceback.format_exc()
                error_message = repr(e)
                additional_info = {
                    "traceback": exception_traceback,
                    "error": error_message,
                }

                return StatusType.CRASHED, cost, 0.0, additional_info  # type: ignore

            if isinstance(rval, tuple):
                result = rval[0]
                additional_run_info = rval[1]
            else:
                result = rval
                additional_run_info = {}

            # get status, cost, time
            if obj.exit_status is pynisher.TimeoutException:
                status = StatusType.TIMEOUT
            elif obj.exit_status is pynisher.MemorylimitException:
                status = StatusType.MEMOUT
            elif obj.exit_status == 0 and result is not None:
                status = StatusType.SUCCESS
                cost = result  # type: ignore # noqa
            else:
                status = StatusType.CRASHED

            runtime = float(obj.wall_clock_time)
        else:
            start_time = time.time()

            # call ta
            try:
                rval = self._call_ta(self._ta, config, obj_kwargs)

                if isinstance(rval, tuple):
                    result = rval[0]
                    additional_run_info = rval[1]
                else:
                    result = rval
                    additional_run_info = {}

                status = StatusType.SUCCESS
                cost = result  # type: ignore
            except Exception as e:
                self.logger.exception(e)
                status = StatusType.CRASHED
                additional_run_info = {}

            runtime = time.time() - start_time

        # Do some sanity checking (for multi objective)
        if len(self.multi_objectives) > 1:
            error = f"Returned costs {cost} does not match the number of objectives {len(self.multi_objectives)}."

            # If dict convert to array
            # Make sure the ordering is correct
            if isinstance(cost, dict):
                ordered_cost = []
                for name in self.multi_objectives:
                    if name not in cost:
                        raise RuntimeError(f"Objective {name} was not found in the returned costs.")

                    ordered_cost.append(cost[name])
                cost = ordered_cost

            if isinstance(cost, list):
                if len(cost) != len(self.multi_objectives):
                    raise RuntimeError(error)

            if isinstance(cost, float):
                raise RuntimeError(error)

        if cost is None or status == StatusType.CRASHED:
            status = StatusType.CRASHED
            cost = self.cost_for_crash

        cost = np.asarray(cost).squeeze().tolist()

        return status, cost, runtime, additional_run_info  # type: ignore

    def _call_ta(
        self,
        obj: Callable,
        config: Configuration,
        obj_kwargs: Dict[str, Union[int, str, float, None]],
    ) -> Union[float, Tuple[float, Dict]]:
        raise NotImplementedError()


class ExecuteTAFuncDict(AbstractTAFunc):
    """Evaluate function for given configuration and resource limit.

    Passes the configuration as a dictionary to the target algorithm. The
    target algorithm needs to implement one of the following signatures:

    * ``target_algorithm(config: Configuration) -> Union[float, Tuple[float, Any]]``
    * ``target_algorithm(config: Configuration, seed: int) -> Union[float, Tuple[float, Any]]``
    * ``target_algorithm(config: Configuration, seed: int, instance: str) -> Union[float, Tuple[float, Any]]``

    The target algorithm can either return a float (the loss), or a tuple
    with the first element being a float and the second being additional run
    information.

    ExecuteTAFuncDict will use inspection to figure out the correct call to
    the target algorithm.

    Parameters
    ----------
    ta : callable
        Function (target algorithm) to be optimized.
    stats : smac.stats.stats.Stats, optional
        Stats object to collect statistics about runtime etc.
    run_obj : str, optional
        Run objective (runtime or quality)
    memory_limit : int, optional
        Memory limit (in MB) that will be applied to the target algorithm.
    par_factor : int, optional
        Penalized average runtime factor. Only used when `run_obj='runtime'`
    use_pynisher: bool, optional
        use pynisher to limit resources;
    """

    def _call_ta(
        self,
        obj: Callable,
        config: Configuration,
        obj_kwargs: Dict[str, Union[int, str, float, None]],
    ) -> Union[float, Tuple[float, Dict]]:

        return obj(config, **obj_kwargs)


class ExecuteTAFuncArray(AbstractTAFunc):
    """Evaluate function for given configuration and resource limit.

    Passes the configuration as an array-like to the target algorithm. The
    target algorithm needs to implement one of the following signatures:

    * ``target_algorithm(config: np.ndarray) -> Union[float, Tuple[float, Any]]``
    * ``target_algorithm(config: np.ndarray, seed: int) -> Union[float, Tuple[float, Any]]``
    * ``target_algorithm(config: np.ndarray, seed: int, instance: str) -> Union[float, Tuple[float, Any]]``

    The target algorithm can either return a float (the loss), or a tuple
    with the first element being a float and the second being additional run
    information.

    ExecuteTAFuncDict will use inspection to figure out the correct call to
    the target algorithm.

    Parameters
    ----------
    ta : callable
        Function (target algorithm) to be optimized.
    stats : smac.stats.stats.Stats, optional
        Stats object to collect statistics about runtime etc.
    run_obj: str, optional
        Run objective (runtime or quality)
    memory_limit : int, optional
        Memory limit (in MB) that will be applied to the target algorithm.
    par_factor: int, optional
        Penalized average runtime factor. Only used when `run_obj='runtime'`
    """

    def _call_ta(
        self,
        obj: Callable,
        config: Configuration,
        obj_kwargs: Dict[str, Union[int, str, float, None]],
    ) -> Union[float, Tuple[float, Dict]]:

        x = np.array([val for _, val in sorted(config.get_dictionary().items())], dtype=float)
        return obj(x, **obj_kwargs)



from abc import abstractmethod

import numpy as np

from smac.optimizer.multi_objective.abstract_multi_objective_algorithm import (
    AbstractMultiObjectiveAlgorithm,
)


class AggregationStrategy(AbstractMultiObjectiveAlgorithm):
    """
    An abstract class to aggregate multi-objective losses to a single objective losses, which can then be utilized
    by the single-objective optimizer.
    """

    @abstractmethod
    def __call__(self, values: np.ndarray) -> float:
        """
        Transform a multi-objective loss to a single loss.

        Parameters
        ----------
            values: np.ndarray[num_evaluations, num_obj].

        Returns
        -------
            cost: float.
        """
        raise NotImplementedError


class MeanAggregationStrategy(AggregationStrategy):
    """
    A class to mean-aggregate multi-objective losses to a single objective losses,
    which can then be utilized by the single-objective optimizer.
    """

    def __call__(self, values: np.ndarray) -> float:
        """
        Transform a multi-objective loss to a single loss.

        Parameters
        ----------
            values (np.ndarray): Normalized values.

        Returns
        -------
            cost (float): Combined cost.
        """
        return np.mean(values, axis=1)



from typing import List, Union

import importlib
import re

import pkg_resources  # type: ignore
from packaging.version import Version

__copyright__ = "Copyright 2021, AutoML.org Freiburg-Hannover"
__license__ = "3-clause BSD"


SUBPATTERN = r"((?P<operation%d>==|>=|>|<)(?P<version%d>(\d+)?(\.[a-zA-Z0-9]+)?(\.\d+)?))"
RE_PATTERN = re.compile(r"^(?P<name>[\w\-]+)%s?(,%s)?$" % (SUBPATTERN % (1, 1), SUBPATTERN % (2, 2)))


def verify_packages(packages: Union[List[str], str]) -> None:
    """Verifies packages. Calls `_verify_packages` as subroutine.

    Parameters
    ----------
    packages : Union[List[str], str]
        Packages to verify.

    Raises
    ------
    ValueError
        If requirements can not be read.
    """
    if not packages:
        return
    if isinstance(packages, str):
        packages = packages.splitlines()

    for package in packages:
        if not package:
            continue

        match = RE_PATTERN.match(package)
        if match:
            name = match.group("name")
            for group_id in range(1, 3):
                if "operation%d" % group_id in match.groupdict():
                    operation = match.group("operation%d" % group_id)
                    version = match.group("version%d" % group_id)
                    _verify_package(name, operation, version)

        else:
            raise ValueError("Unable to read requirement: %s" % package)


def _verify_package(name: str, operation: str, version: str) -> None:
    """Verifies a package.

    Parameters
    ----------
    name : str
        Name of the package.
    operation : str
        Operation of the package.
    version : str
        Version of the package.

    Raises
    ------
    MissingPackageError
        If package is missing.
    NotImplementedError
        If operation is not implemented.
    IncorrectPackageVersionError
        If package version is incorrect.
    """
    try:
        distribution = pkg_resources.get_distribution(name)
        installed_version = Version(distribution.version)
    except pkg_resources.DistributionNotFound:
        try:
            module = importlib.import_module(name)
            installed_version = Version(module.__version__)  # type: ignore[attr-defined] # noqa F821
        except ImportError:
            raise MissingPackageError(name)

    if not operation:
        return

    # pkg_resources.get_distribution can (not) find a version depending on how the package was built
    # if we get version 0.0.0 we fallback to the module's version
    if installed_version == Version("0.0.0"):
        module = importlib.import_module(name)
        installed_version = Version(module.__version__)

    required_version = Version(version)

    if operation == "==":
        check = required_version == installed_version
    elif operation == ">":
        check = installed_version > required_version
    elif operation == "<":
        check = installed_version < required_version
    elif operation == ">=":
        check = installed_version > required_version or installed_version == required_version
    elif operation == "<=":
        check = installed_version < required_version or installed_version == required_version
    else:
        raise NotImplementedError("operation '%s' is not supported" % operation)
    if not check:
        raise IncorrectPackageVersionError(name, installed_version, operation, required_version)


class MissingPackageError(Exception):
    error_message = "Mandatory package '{name}' not found!"

    def __init__(self, package_name: str) -> None:
        self.package_name = package_name
        super(MissingPackageError, self).__init__(self.error_message.format(name=package_name))


class IncorrectPackageVersionError(Exception):
    error_message = "'{name} {installed_version}' version mismatch ({operation}{required_version})"

    def __init__(
        self,
        package_name: str,
        installed_version: Version,
        operation: str,
        required_version: Version,
    ) -> None:
        self.package_name = package_name
        self.installed_version = installed_version
        self.operation = operation
        self.required_version = required_version
        message = self.error_message.format(
            name=package_name,
            installed_version=installed_version,
            operation=operation,
            required_version=required_version,
        )
        super(IncorrectPackageVersionError, self).__init__(message)



from typing import Any

from smac.epm.rf_with_instances import RandomForestWithInstances
from smac.facade.smac_ac_facade import SMAC4AC
from smac.initial_design.sobol_design import SobolDesign
from smac.optimizer.acquisition import LogEI
from smac.runhistory.runhistory2epm import RunHistory2EPM4LogScaledCost

__author__ = "Marius Lindauer"
__copyright__ = "Copyright 2018, ML4AAD"
__license__ = "3-clause BSD"


class SMAC4HPO(SMAC4AC):
    """Facade to use SMAC for hyperparameter optimization.

    see smac.facade.smac_Facade for API
    This facade overwrites options available via the SMAC facade

    See Also
    --------
    :class:`~smac.facade.smac_ac_facade.SMAC4AC` for documentation of parameters.

    Attributes
    ----------
    logger
    stats : Stats
    solver : SMBO
    runhistory : RunHistory
        List with information about previous runs
    trajectory : list
        List of all incumbents
    """

    def __init__(self, **kwargs: Any):
        scenario = kwargs["scenario"]

        kwargs["initial_design"] = kwargs.get("initial_design", SobolDesign)
        if len(scenario.cs.get_hyperparameters()) > 21201 and kwargs["initial_design"] is SobolDesign:
            raise ValueError(
                'The default initial design "Sobol sequence" can only handle up to 21201 dimensions. '
                'Please use a different initial design, such as "the Latin Hypercube design".',
            )

        init_kwargs = kwargs.get("initial_design_kwargs", dict())
        init_kwargs["n_configs_x_params"] = init_kwargs.get("n_configs_x_params", 10)
        init_kwargs["max_config_fracs"] = init_kwargs.get("max_config_fracs", 0.25)
        kwargs["initial_design_kwargs"] = init_kwargs

        # Intensification parameters - which intensifier to use and respective parameters
        intensifier_kwargs = kwargs.get("intensifier_kwargs", dict())
        intensifier_kwargs["min_chall"] = 1
        kwargs["intensifier_kwargs"] = intensifier_kwargs
        scenario.intensification_percentage = 1e-10

        if kwargs.get("model") is None:
            model_class = RandomForestWithInstances
            kwargs["model"] = model_class

            # == static RF settings
            model_kwargs = kwargs.get("model_kwargs", dict())
            model_kwargs["num_trees"] = model_kwargs.get("num_trees", 10)
            model_kwargs["do_bootstrapping"] = model_kwargs.get("do_bootstrapping", True)
            model_kwargs["ratio_features"] = model_kwargs.get("ratio_features", 1.0)
            model_kwargs["min_samples_split"] = model_kwargs.get("min_samples_split", 2)
            model_kwargs["min_samples_leaf"] = model_kwargs.get("min_samples_leaf", 1)
            model_kwargs["log_y"] = model_kwargs.get("log_y", True)
            kwargs["model_kwargs"] = model_kwargs

        # == Acquisition function
        kwargs["acquisition_function"] = kwargs.get("acquisition_function", LogEI)
        kwargs["runhistory2epm"] = kwargs.get("runhistory2epm", RunHistory2EPM4LogScaledCost)

        # assumes random chooser for random configs
        random_config_chooser_kwargs = kwargs.get("random_configuration_chooser_kwargs", dict())
        random_config_chooser_kwargs["prob"] = random_config_chooser_kwargs.get("prob", 0.2)
        kwargs["random_configuration_chooser_kwargs"] = random_config_chooser_kwargs

        # better improve acquisition function optimization
        # 1. increase number of sls iterations
        acquisition_function_optimizer_kwargs = kwargs.get("acquisition_function_optimizer_kwargs", dict())
        acquisition_function_optimizer_kwargs["n_sls_iterations"] = 10
        kwargs["acquisition_function_optimizer_kwargs"] = acquisition_function_optimizer_kwargs

        super().__init__(**kwargs)
        self.logger.info(self.__class__)

        # better improve acquisition function optimization
        # 2. more randomly sampled configurations
        self.solver.scenario.acq_opt_challengers = 10000  # type: ignore[attr-defined] # noqa F821

        # activate predict incumbent
        self.solver.epm_chooser.predict_x_best = True



from typing import Any, Callable, Dict, Iterable, List, Mapping, Optional, Tuple, Union

import logging

import numpy as np
from ConfigSpace.hyperparameters import UniformFloatHyperparameter  # type: ignore

from smac.configspace import Configuration, ConfigurationSpace
from smac.facade.smac_hpo_facade import SMAC4HPO
from smac.runhistory.runhistory import RunKey
from smac.scenario.scenario import Scenario
from smac.tae.execute_func import ExecuteTAFuncArray

__author__ = "Marius Lindauer, Matthias Feurer"
__copyright__ = "Copyright 2016, ML4AAD"
__license__ = "3-clause BSD"


def fmin_smac(
    func: Callable,
    x0: List[float],
    bounds: List[Iterable[float]],
    maxfun: int = -1,
    rng: Optional[Union[np.random.RandomState, int]] = None,
    scenario_args: Optional[Mapping[str, Any]] = None,
    tae_runner_kwargs: Optional[Dict[str, Any]] = None,
    **kwargs: Any,
) -> Tuple[Configuration, Union[np.ndarray, float], SMAC4HPO]:
    """Minimize a function func using the SMAC4HPO facade (i.e., a modified version of SMAC). This
    function is a convenience wrapper for the SMAC4HPO class.

    Parameters
    ----------
    func : Callable
        Function to minimize.
    x0 : List[float]
        Initial guess/default configuration.
    bounds : List[List[float]]
        ``(min, max)`` pairs for each element in ``x``, defining the bound on
        that parameters.
    maxfun : int, optional
        Maximum number of function evaluations.
    rng : np.random.RandomState, optional
            Random number generator used by SMAC.
    scenario_args: Mapping[str,Any]
        Arguments passed to the scenario
        See smac.scenario.scenario.Scenario
    **kwargs:
        Arguments passed to the optimizer class
        See ~smac.facade.smac_facade.SMAC

    Returns
    -------
    x : list
        Estimated position of the minimum.
    f : Union[np.ndarray, float]
        Value of `func` at the minimum. Depending on the scenario_args, it could be a scalar value
        (for single objective problems) or a np.ndarray (for multi objective problems).
    s : :class:`smac.facade.smac_hpo_facade.SMAC4HPO`
        SMAC objects which enables the user to get
        e.g., the trajectory and runhistory.
    """
    # create configuration space
    cs = ConfigurationSpace()

    # Adjust zero padding
    tmplt = "x{0:0" + str(len(str(len(bounds)))) + "d}"

    for idx, (lower_bound, upper_bound) in enumerate(bounds):
        parameter = UniformFloatHyperparameter(
            name=tmplt.format(idx + 1),
            lower=lower_bound,
            upper=upper_bound,
            default_value=x0[idx],
        )
        cs.add_hyperparameter(parameter)

    # create scenario
    scenario_dict = {
        "run_obj": "quality",
        "cs": cs,
        "deterministic": True,
        "initial_incumbent": "DEFAULT",
    }

    if scenario_args is not None:
        scenario_dict.update(scenario_args)

    if maxfun > 0:
        scenario_dict["runcount_limit"] = maxfun
    scenario = Scenario(scenario_dict)

    # Handle optional tae  arguments
    if tae_runner_kwargs is not None:
        if "ta" not in tae_runner_kwargs:
            tae_runner_kwargs.update({"ta": func})
    else:
        tae_runner_kwargs = {"ta": func}

    smac = SMAC4HPO(
        scenario=scenario,
        tae_runner=ExecuteTAFuncArray,
        tae_runner_kwargs=tae_runner_kwargs,
        rng=rng,
        **kwargs,
    )

    smac.logger = logging.getLogger(smac.__module__ + "." + smac.__class__.__name__)
    incumbent = smac.optimize()
    config_id = smac.solver.runhistory.config_ids[incumbent]
    run_key = RunKey(config_id, None, 0)
    incumbent_performance = smac.solver.runhistory.data[run_key]
    incumbent = np.array([incumbent[tmplt.format(idx + 1)] for idx in range(len(bounds))], dtype=float)

    return incumbent, incumbent_performance.cost, smac



from abc import abstractmethod
from typing import Any, Dict, List, Optional, Tuple

import numpy as np

from smac.configspace import ConfigurationSpace
from smac.epm.base_epm import AbstractEPM

__copyright__ = "Copyright 2021, AutoML.org Freiburg-Hannover"
__license__ = "3-clause BSD"


class UncorrelatedMultiObjectiveModel(AbstractEPM):
    """Wrapper for the surrogate models to predict multiple targets.

    Only a list with the target names and the types array for the
    underlying model are mandatory. All other hyperparameters to
    model can be passed via kwargs. Consult the documentation of
    the corresponding model for the hyperparameters and their meanings.

    Parameters
    ----------
    target_names : list
        List of str, each entry is the name of one target dimension. Length
        of the list will be ``n_objectives``.
    types : List[int]
        Specifies the number of categorical values of an input dimension where
        the i-th entry corresponds to the i-th input dimension. Let's say we
        have 2 dimension where the first dimension consists of 3 different
        categorical choices and the second dimension is continuous than we
        have to pass [3, 0]. Note that we count starting from 0.
    bounds : List[Tuple[float, float]]
        bounds of input dimensions: (lower, uppper) for continuous dims; (n_cat, np.nan)
        for categorical dims
    instance_features : np.ndarray (I, K)
        Contains the K dimensional instance features of I different instances
    pca_components : float
        Number of components to keep when using PCA to reduce dimensionality of instance features.
        Requires to set n_feats (> pca_dims).
    model_kwargs: Optional[Dict[str, Any]]:
        arguments for initialing estimators

    Attributes
    ----------
    target_names:
        target names
    num_targets: int
        number of targets
    estimators: List[AbstractEPM]
        a list of estimators predicting different target values
    """

    def __init__(
        self,
        target_names: List[str],
        configspace: ConfigurationSpace,
        types: List[int],
        bounds: List[Tuple[float, float]],
        seed: int,
        instance_features: Optional[np.ndarray] = None,
        pca_components: Optional[int] = None,
        model_kwargs: Optional[Dict[str, Any]] = None,
    ) -> None:
        super().__init__(
            configspace=configspace,
            bounds=bounds,
            types=types,
            seed=seed,
            instance_features=instance_features,
            pca_components=pca_components,
        )
        if model_kwargs is None:
            model_kwargs = {}
        self.target_names = target_names
        self.num_targets = len(self.target_names)
        self.estimators: List[AbstractEPM] = self.construct_estimators(configspace, types, bounds, model_kwargs)

    @abstractmethod
    def construct_estimators(
        self,
        configspace: ConfigurationSpace,
        types: List[int],
        bounds: List[Tuple[float, float]],
        model_kwargs: Dict[str, Any],
    ) -> List[AbstractEPM]:
        """
        Construct a list of estimators. The number of the estimators equals 'self.num_targets'
        Parameters
        ----------
        configspace : ConfigurationSpace
            Configuration space to tune for.
        types : List[int]
            Specifies the number of categorical values of an input dimension where
            the i-th entry corresponds to the i-th input dimension. Let's say we
            have 2 dimension where the first dimension consists of 3 different
            categorical choices and the second dimension is continuous than we
            have to pass [3, 0]. Note that we count starting from 0.
        bounds : List[Tuple[float, float]]
            bounds of input dimensions: (lower, uppper) for continuous dims; (n_cat, np.nan) for categorical dims
        model_kwargs : Dict[str, Any]
            model kwargs for initializing models
        Returns
        -------
        estimators: List[AbstractEPM]
            A list of estimators
        """
        raise NotImplementedError

    def _train(self, X: np.ndarray, Y: np.ndarray) -> "UncorrelatedMultiObjectiveModel":
        """Trains the models on X and y.

        Parameters
        ----------
        X : np.ndarray [n_samples, n_features (config + instance features)]
            Input data points.
        Y : np.ndarray [n_samples, n_objectives]
            The corresponding target values. n_objectives must match the
            number of target names specified in the constructor.

        Returns
        -------
        self
        """
        if len(self.estimators) == 0:
            raise ValueError("The list of estimators for this model is empty!")
        for i, estimator in enumerate(self.estimators):
            estimator.train(X, Y[:, i])

        return self

    def _predict(self, X: np.ndarray, cov_return_type: Optional[str] = "diagonal_cov") -> Tuple[np.ndarray, np.ndarray]:
        """Predict means and variances for given X.

        Parameters
        ----------
        X : np.ndarray of shape = [n_samples, n_features (config + instance
        features)]
        cov_return_type: Optional[str]
            Specifies what to return along with the mean. Refer ``predict()`` for more information.

        Returns
        -------
        means : np.ndarray of shape = [n_samples, n_objectives]
            Predictive mean
        vars : np.ndarray  of shape = [n_samples, n_objectives]
            Predictive variance
        """
        if cov_return_type != "diagonal_cov":
            raise ValueError("'cov_return_type' can only take 'diagonal_cov' for this model")

        mean = np.zeros((X.shape[0], self.num_targets))
        var = np.zeros((X.shape[0], self.num_targets))
        for i, estimator in enumerate(self.estimators):
            m, v = estimator.predict(X)
            assert v is not None  # please mypy
            mean[:, i] = m.flatten()
            var[:, i] = v.flatten()
        return mean, var

    def predict_marginalized_over_instances(self, X: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """Predict mean and variance marginalized over all instances.

        Returns the predictive mean and variance marginalised over all
        instances for a set of configurations.

        Parameters
        ----------
        X : np.ndarray of shape = [n_features (config), ]

        Returns
        -------
        means : np.ndarray of shape = [n_samples, n_objectives]
            Predictive mean
        vars : np.ndarray  of shape = [n_samples, n_objectives]
            Predictive variance
        """
        mean = np.zeros((X.shape[0], self.num_targets))
        var = np.zeros((X.shape[0], self.num_targets))
        for i, estimator in enumerate(self.estimators):
            m, v = estimator.predict_marginalized_over_instances(X)
            mean[:, i] = m.flatten()
            var[:, i] = v.flatten()
        return mean, var



from typing import Optional

import numpy as np

from smac.optimizer.multi_objective.aggregation_strategy import AggregationStrategy


class ParEGO(AggregationStrategy):
    def __init__(
        self,
        num_obj: int,
        rng: Optional[np.random.RandomState] = None,
        rho: float = 0.05,
    ):
        super(ParEGO, self).__init__(num_obj=num_obj, rng=rng)
        self.rho = rho

    def __call__(self, values: np.ndarray) -> float:
        """
        Transform a multi-objective loss to a single loss.

        Parameters
        ----------
            values (np.ndarray): Normalized values.

        Returns
        -------
            cost (float): Combined cost.
        """
        # Then we have to compute the weight
        theta = self.rng.rand(self.num_obj)

        # Normalize st all theta values sum up to 1
        theta = theta / (np.sum(theta) + 1e-10)

        # Weight the values
        theta_f = theta * values

        return np.max(theta_f, axis=1) + self.rho * np.sum(theta_f, axis=1)



from typing import Any, Dict, Iterable, List, Union

import logging

import numpy as np

__copyright__ = "Copyright 2021, AutoML.org Freiburg-Hannover"
__license__ = "3-clause BSD"


class PickableLoggerAdapter(object):
    def __init__(self, name: str) -> None:
        self.name = name
        self.logger = logging.getLogger(self.name)

    def __getstate__(self) -> Dict[str, str]:
        """
        Method is called when pickle dumps an object.

        Returns
        -------
        Dictionary, representing the object state to be pickled. Ignores
        the self.logger field and only returns the logger name.
        """
        return {"name": self.name}

    def __setstate__(self, state: Dict[str, Any]) -> None:
        """
        Method is called when pickle loads an object. Retrieves the name and
        creates a logger.

        Parameters
        ----------
        state - dictionary, containing the logger name.
        """
        self.name = state["name"]
        self.logger = logging.getLogger(self.name)

    def debug(self, msg, *args, **kwargs):  # type: ignore[no-untyped-def] # noqa F821
        """Debug method."""
        self.logger.debug(msg, *args, **kwargs)

    def info(self, msg, *args, **kwargs):  # type: ignore[no-untyped-def] # noqa F821
        """Info method."""
        self.logger.info(msg, *args, **kwargs)

    def warning(self, msg, *args, **kwargs):  # type: ignore[no-untyped-def] # noqa F821
        """Warning method."""
        self.logger.warning(msg, *args, **kwargs)

    def error(self, msg, *args, **kwargs):  # type: ignore[no-untyped-def] # noqa F821
        """Error method."""
        self.logger.error(msg, *args, **kwargs)

    def exception(self, msg, *args, **kwargs):  # type: ignore[no-untyped-def] # noqa F821
        """Exception method."""
        self.logger.exception(msg, *args, **kwargs)

    def critical(self, msg, *args, **kwargs):  # type: ignore[no-untyped-def] # noqa F821
        """Critical method."""
        self.logger.critical(msg, *args, **kwargs)

    def log(self, level, msg, *args, **kwargs):  # type: ignore[no-untyped-def] # noqa F821
        """Log method."""
        self.logger.log(level, msg, *args, **kwargs)

    def isEnabledFor(self, level):  # type: ignore[no-untyped-def] # noqa F821
        """Check if logger is enabled for a given level."""
        return self.logger.isEnabledFor(level)


def format_array(
    inputs: Union[str, int, float, np.ndarray, list], format_vals: bool = True
) -> Union[float, List[float]]:
    """Transform a numpy array to a list of format so that it can be printed by logger. If the list
    holds one element only, then a formatted string is returned.

    Parameters
    ----------
        inputs: np.ndarray or list.
            inputs value, could be anything serializable or a np array
        format_vals: bool.
            if the items in list are formatted values

    Returns
    -------
        result: float or list of floats.
    """
    if isinstance(inputs, np.ndarray):
        inputs = inputs.tolist()

    if not isinstance(inputs, Iterable):
        inputs = [inputs]

    formatted_list = []
    for item in inputs:
        item = float(item)
        if format_vals:
            item = np.round(item, 4)

        formatted_list.append(item)

    if len(formatted_list) == 1:
        return formatted_list[0]

    return formatted_list



from abc import ABC, abstractmethod
from typing import Optional

import logging

import numpy as np

__author__ = "Aaron Kimmig"
__copyright__ = "Copyright 2015, ML4AAD"
__license__ = "3-clause BSD"
__maintainer__ = "Aaron Kimmig"
__email__ = "kimmiga@cs.uni-freiburg.de"
__version__ = "0.0.1"


class RandomConfigurationChooser(ABC):
    """Abstract base of helper classes to configure interleaving of random configurations in a list
    of challengers.
    """

    def __init__(self, rng: Optional[np.random.RandomState] = None):
        self.rng = rng or np.random.RandomState(seed=0)

    @abstractmethod
    def next_smbo_iteration(self) -> None:
        """Indicate beginning of next SMBO iteration."""
        pass

    @abstractmethod
    def check(self, iteration: int) -> bool:
        """Check if the next configuration should be at random."""
        pass


class ChooserNoCoolDown(RandomConfigurationChooser):
    """Interleave a random configuration after a constant number of configurations found by Bayesian
    optimization.

    Parameters
    ----------
    modulus : float
        Every modulus-th configuration will be at random.
    """

    def __init__(self, rng: Optional[np.random.RandomState] = None, modulus: float = 2.0):
        super().__init__(rng)

        self.logger = logging.getLogger(self.__module__ + "." + self.__class__.__name__)
        if modulus <= 1.0:
            self.logger.warning("Using SMAC with random configurations only." "ROAR is the better choice for this.")
        self.modulus = modulus

    def next_smbo_iteration(self) -> None:
        """Does nothing."""
        ...

    def check(self, iteration: int) -> bool:
        """Checks if the next configuration should be at random."""
        return iteration % self.modulus < 1


class ChooserLinearCoolDown(RandomConfigurationChooser):
    """Interleave a random configuration, decreasing the fraction of random configurations over
    time.

    Parameters
    ----------
    start_modulus : float
       Initially, every modulus-th configuration will be at random
    modulus_increment : float
       Increase modulus by this amount in every iteration
    end_modulus : float
       Highest modulus used in the chooser. If the value is reached before the optimization is over, it is not
       further increased. If it is not reached before the optimization is over, there will be no adjustment to make
       sure that the ``end_modulus`` is reached.
    """

    def __init__(
        self,
        rng: Optional[np.random.RandomState] = None,
        start_modulus: float = 2.0,
        modulus_increment: float = 0.3,
        end_modulus: float = np.inf,
    ):
        super().__init__(rng)

        self.logger = logging.getLogger(self.__module__ + "." + self.__class__.__name__)
        if start_modulus <= 1.0 and modulus_increment <= 0.0:
            self.logger.warning("Using SMAC with random configurations only. ROAR is the better choice for this.")
        self.modulus = start_modulus
        self.modulus_increment = modulus_increment
        self.end_modulus = end_modulus
        self.last_iteration = 0

    def next_smbo_iteration(self) -> None:
        """Change modulus."""
        self.modulus += self.modulus_increment
        self.modulus = min(self.modulus, self.end_modulus)
        self.last_iteration = 0

    def check(self, iteration: int) -> bool:
        """Check if the next configuration should be interleaved based on modulus."""
        if (iteration - self.last_iteration) % self.modulus < 1:
            self.last_iteration = iteration
            return True
        else:
            return False


class ChooserProb(RandomConfigurationChooser):
    """Interleave a random configuration according to a given probability.

    Parameters
    ----------
    prob : float
        Probility of a random configuration
    rng : np.random.RandomState
        Random state
    """

    def __init__(self, rng: Optional[np.random.RandomState], prob: float):
        super().__init__(rng)
        self.prob = prob

    def next_smbo_iteration(self) -> None:
        """Does nothing."""
        ...

    def check(self, iteration: int) -> bool:
        """Check if the next configuration should be at random."""
        if self.rng.rand() < self.prob:
            return True
        else:
            return False


class ChooserProbCoolDown(RandomConfigurationChooser):
    """Interleave a random configuration according to a given probability which is decreased over
    time.

    Parameters
    ----------
    prob : float
        Probility of a random configuration
    cool_down_fac : float
        Multiply the ``prob`` by ``cool_down_fac`` in each iteration
    rng : np.random.RandomState
        Random state
    """

    def __init__(self, rng: Optional[np.random.RandomState], prob: float, cool_down_fac: float):
        super().__init__(rng)
        self.prob = prob
        self.cool_down_fac = cool_down_fac

    def next_smbo_iteration(self) -> None:
        """Set the probability to the current value multiplied by the `cool_down_fac`."""
        self.prob *= self.cool_down_fac

    def check(self, iteration: int) -> bool:
        """Check if the next configuration should be at random."""
        if self.rng.rand() < self.prob:
            return True
        else:
            return False


class ChooserCosineAnnealing(RandomConfigurationChooser):
    """Interleave a random configuration according to a given probability which is decreased
    according to a cosine annealing schedule.

    Parameters
    ----------
    prob_max : float
        Initial probility of a random configuration
    prob_min : float
        Lowest probility of a random configuration
    restart_iteration : int
        Restart the annealing schedule every ``restart_iteration`` iterations.
    rng : np.random.RandomState
        Random state
    """

    def __init__(
        self,
        rng: Optional[np.random.RandomState],
        prob_max: float,
        prob_min: float,
        restart_iteration: int,
    ):
        super().__init__(rng)
        self.logger = logging.getLogger(self.__module__ + "." + self.__class__.__name__)
        self.prob_max = prob_max
        self.prob_min = prob_min
        self.restart_iteration = restart_iteration
        self.iteration = 0
        self.prob = prob_max

    def next_smbo_iteration(self) -> None:
        """Set `self.prob` and increases the iteration counter."""
        self.prob = self.prob_min + (
            0.5 * (self.prob_max - self.prob_min) * (1 + np.cos(self.iteration * np.pi / self.restart_iteration))
        )
        self.logger.error("Probability for random configs: %f" % self.prob)
        self.iteration += 1
        if self.iteration > self.restart_iteration:
            self.iteration = 0
            self.logger.error("Perform restart in next iteration!")

    def check(self, iteration: int) -> bool:
        """Check if a random configuration should be interleaved."""
        if self.rng.rand() < self.prob:
            self.logger.error("Random Config")
            return True
        else:
            self.logger.error("Acq Config")
            return False



from typing import Any, Dict, List, Optional, Tuple

import numpy as np

from smac.configspace import ConfigurationSpace
from smac.configspace import json as pcs_json
from smac.configspace import pcs, pcs_new
from smac.utils.logging import PickableLoggerAdapter

__author__ = "Marius Lindauer"
__copyright__ = "Copyright 2015, ML4AAD"
__license__ = "3-clause BSD"
__maintainer__ = "Marius Lindauer"
__email__ = "lindauer@cs.uni-freiburg.de"
__version__ = "0.0.1"


INSTANCE_TYPE = List[List[str]]
INSTANCE_FEATURES_TYPE = Tuple[List[str], Dict[str, np.ndarray]]


class InputReader(object):
    """Reading all input files for SMAC (scenario file, instance files, ...)

    Note: Most of this code was taken from the pysmac repository.
          We copy it here because we don't want smac3 to depend
          on an earlier version!
    """

    def __init__(self) -> None:
        pass

    def read_scenario_file(self, fn: str) -> Dict[str, Any]:
        """Encapsulates read_scenario_file of pysmac

        Parameters
        ----------
            fn: string
                 File name of scenario file

        Returns
        -------
        dict : dictionary
            (key, value) pairs are (variable name, variable value)
        """
        # translate the difference option names to a canonical name
        # kept for backwards-compatibility
        scenario_option_names = {
            "algo-exec": "algo",
            "algoExec": "algo",
            "algo": "algo",
            "algo-exec-dir": "execdir",
            "exec-dir": "execdir",
            "execDir": "execdir",
            "execdir": "execdir",
            "algo-deterministic": "deterministic",
            "deterministic": "deterministic",
            "paramFile": "paramfile",
            "pcs-file": "paramfile",
            "param-file": "paramfile",
            "paramfile": "paramfile",
            "run-obj": "run_obj",
            "run-objective": "run_obj",
            "runObj": "run_obj",
            "run_obj": "run_obj",
            "overall_obj": "overall_obj",
            "intra-obj": "overall_obj",
            "intra-instance-obj": "overall_obj",
            "overall-obj": "overall_obj",
            "intraInstanceObj": "overall_obj",
            "overallObj": "overall_obj",
            "intra_instance_obj": "overall_obj",
            "cost-for-crash": "cost_for_crash",
            "cost_for_crash": "cost_for_crash",
            "algo-cutoff-time": "cutoff_time",
            "target-run-cputime-limit": "cutoff_time",
            "target_run_cputime_limit": "cutoff_time",
            "cutoff-time": "cutoff_time",
            "cutoffTime": "cutoff_time",
            "cutoff_time": "cutoff_time",
            "memory-limit": "memory_limit",
            "memory_limit": "memory_limit",
            "cputime-limit": "tuner_timeout",
            "cputime_limit": "tuner_timeout",
            "tunertime-limit": "tuner_timeout",
            "tuner-timeout": "tuner_timeout",
            "tunerTimeout": "tuner_timeout",
            "tuner_timeout": "tuner_timeout",
            "wallclock-limit": "wallclock_limit",
            "runtime-limit": "wallclock_limit",
            "runtimeLimit": "wallclock_limit",
            "wallClockLimit": "wallclock_limit",
            "wallclock_limit": "wallclock_limit",
            "output-dir": "output_dir",
            "outputDirectory": "output_dir",
            "outdir": "output_dir",
            "output_dir": "output_dir",
            "instances": "instance_file",
            "instance-file": "instance_file",
            "instance-dir": "instance_file",
            "instanceFile": "instance_file",
            "instance_file": "instance_file",
            "i": "instance_file",
            "instance_seed_file": "instance_file",
            "test-instances": "test_instance_file",
            "test-instance-file": "test_instance_file",
            "test-instance-dir": "test_instance_file",
            "testInstanceFile": "test_instance_file",
            "test_instance_file": "test_instance_file",
            "test_instance_seed_file": "test_instance_file",
            "feature-file": "feature_file",
            "instanceFeatureFile": "feature_file",
            "feature_file": "feature_file",
            "runcount-limit": "runcount_limit",
            "runcount_limit": "runcount_limit",
            "totalNumRunsLimit": "runcount_limit",
            "numRunsLimit": "runcount_limit",
            "numberOfRunsLimit": "runcount_limit",
            "initial-incumbent": "initial_incumbent",
            "initial_incumbent": "initial_incumbent",
        }

        scenario_dict = {}
        with open(fn, "r") as fh:
            for line in fh:
                line = line.replace("\n", "").strip(" ")
                # remove comments
                if line.find("#") > -1:
                    line = line[: line.find("#")]

                # skip empty lines
                if line == "":
                    continue
                if "=" in line:
                    tmp = line.split("=")
                    tmp = [" ".join(s.split()) for s in tmp]
                else:
                    tmp = line.split()
                scenario_dict[scenario_option_names.get(tmp[0], tmp[0])] = " ".join(tmp[1:])
        return scenario_dict

    def read_instance_file(self, fn: str) -> INSTANCE_TYPE:
        """Encapsulates read_instances_file of pysmac

        Parameters
        ----------
            fn: string
                 File name of instance file

        Returns
        -------
            instances: list
                Each element is a list where the first element is the
                instance name followed by additional
                information for the specific instance.
        """
        with open(fn, "r") as fh:
            instance_names = fh.readlines()
        return [s.strip().split() for s in instance_names]

    def read_instance_features_file(
        self,
        fn: str,
    ) -> INSTANCE_FEATURES_TYPE:
        """Encapsulates read_instances_file of pysmac

        Parameters
        ----------
            fn: string
                File name of instance feature file

        Returns
        -------
            features: tuple
                first entry is a list of the feature names,
                second one is a dict with 'instance name' -
                'numpy array containing the features' key-value pairs
        """
        instances = {}
        with open(fn, "r") as fh:
            lines = fh.readlines()
            for line in lines[1:]:
                tmp = line.strip().split(",")
                instances[tmp[0]] = np.array(tmp[1:], dtype=np.double)
        return [f.strip() for f in lines[0].rstrip("\n").split(",")[1:]], instances  # type: ignore

    @staticmethod
    def read_pcs_file(fn: str, logger: Optional[PickableLoggerAdapter] = None) -> ConfigurationSpace:
        """Encapsulates generating configuration space object from file.

        Automatically detects whether the cs is saved in json, pcs or pcs_new.

        Parameters
        ----------
            fn: string
                 File name of pcs file

        Returns
        -------
            ConfigSpace: ConfigSpace
        """
        # Three possible formats: json, pcs and pcs_new. We prefer json.
        with open(fn) as fp:
            if fn.endswith(".json"):
                cs = pcs_json.read(fp.read())
                if logger:
                    logger.debug("Loading pcs as json from: %s", fn)
            else:
                pcs_str = fp.readlines()
                try:
                    cs = pcs.read(pcs_str)
                except NotImplementedError:
                    if logger:
                        logger.debug("Could not parse pcs file with old format; trying new format ...")
                    cs = pcs_new.read(pcs_str)
        return cs



from typing import Callable, Dict, List, Optional, Tuple, Union

from smac.configspace import Configuration
from smac.runhistory.runhistory import RunInfo, RunValue
from smac.stats.stats import Stats
from smac.tae import StatusType
from smac.tae.base import BaseRunner
from smac.utils.constants import MAXINT

__copyright__ = "Copyright 2021, AutoML.org Freiburg-Hannover"
__license__ = "3-clause BSD"


class SerialRunner(BaseRunner):
    """Interface to submit and collect a job in a serial fashion.

    It dictates what a worker should do to convert a
    configuration/instance/seed to a result.

    This class is expected to be extended via the implementation of
    a run() method for the desired task.

    Attributes
    ----------
    results
    ta
    stats
    run_obj
    par_factor
    cost_for_crash
    abort_i_first_run_crash

    Parameters
    ----------
    ta : list
        target algorithm command line as list of arguments
    stats: Stats()
         stats object to collect statistics about runtime and so on
    multi_objectives: List[str]
        names of the objectives, by default it is a single objective parameter "cost"
    run_obj: str
        run objective of SMAC
    par_factor: int
        penalization factor
    cost_for_crash : float
        cost that is used in case of crashed runs (including runs
        that returned NaN or inf)
    abort_on_first_run_crash: bool
        if true and first run crashes, raise FirstRunCrashedException
    """

    def __init__(
        self,
        ta: Union[List[str], Callable],
        stats: Stats,
        multi_objectives: List[str] = ["cost"],
        run_obj: str = "runtime",
        par_factor: int = 1,
        cost_for_crash: Union[float, List[float]] = float(MAXINT),
        abort_on_first_run_crash: bool = True,
    ):
        super(SerialRunner, self).__init__(
            ta=ta,
            stats=stats,
            multi_objectives=multi_objectives,
            run_obj=run_obj,
            par_factor=par_factor,
            cost_for_crash=cost_for_crash,
            abort_on_first_run_crash=abort_on_first_run_crash,
        )

    def submit_run(self, run_info: RunInfo) -> None:
        """This function submits a run_info object in a serial fashion.

        As there is a single worker for this task, this
        interface can be considered a wrapper over the run()
        method.

        Both result/exceptions can be completely determined in this
        step so both lists are properly filled.

        Parameters
        ----------
        run_info: RunInfo
            An object containing the configuration and the necessary data to run it
        """
        self.results.append(self.run_wrapper(run_info))

    def get_finished_runs(self) -> List[Tuple[RunInfo, RunValue]]:
        """This method returns any finished configuration, and returns a list with the results of
        exercising the configurations. This class keeps populating results to self.results until a
        call to get_finished runs is done. In this case, the self.results list is emptied and all
        RunValues produced by running self.run() are returned.

        Returns
        -------
            List[RunInfo, RunValue]: A list of RunInfo/RunValues pairs
            a submitted configuration
        """
        results_list = []
        while self.results:
            results_list.append(self.results.pop())
        return results_list

    def wait(self) -> None:
        """SMBO/intensifier might need to wait for runs to finish before making a decision.

        For serial runs, no wait is needed as the result is immediately available.
        """
        # There is no need to wait in serial runs.
        # When launching a run via submit, as the serial run
        # uses the same process to run, the result is always available
        # immediately after. This method implements is just an implementation of the
        # abstract method via a simple return, again, because there is
        # no need to wait (as in distributed runs)
        return

    def pending_runs(self) -> bool:
        """Whether or not there are configs still running.

        Generally if the runner is serial, launching a run instantly returns it's result. On
        parallel runners, there might be pending configurations to complete.
        """
        # No pending runs in a serial run. Execution is blocking
        return False

    def run(
        self,
        config: Configuration,
        instance: str,
        cutoff: Optional[float] = None,
        seed: int = 12345,
        budget: Optional[float] = None,
        instance_specific: str = "0",
    ) -> Tuple[StatusType, float, float, Dict]:
        """Runs target algorithm <self.ta> with configuration <config> on instance <instance> with
        instance specifics.

        <specifics> for at most.

        <cutoff> seconds and random seed <seed>

        This method exemplifies how to defined the run() method

        Parameters
        ----------
            config : Configuration
                dictionary param -> value
            instance : string
                problem instance
            cutoff : float, optional
                Wallclock time limit of the target algorithm. If no value is
                provided no limit will be enforced.
            seed : int
                random seed
            budget : float, optional
                A positive, real-valued number representing an arbitrary limit to the target
                algorithm. Handled by the target algorithm internally
            instance_specific: str
                instance specific information (e.g., domain file or solution)

        Returns
        -------
            status: enum of StatusType (int)
                {SUCCESS, TIMEOUT, CRASHED, ABORT}
            cost: float
                cost/regret/quality (float) (None, if not returned by TA)
            runtime: float
                runtime (None if not returned by TA)
            additional_info: dict
                all further additional run information
        """
        pass

    def num_workers(self) -> int:
        """Total number of workers available."""
        # Any serial runner supports only 1 worker
        return 1



from typing import Iterator, List, Optional, Tuple

import logging

import numpy as np

from smac.configspace import Configuration
from smac.configspace.util import convert_configurations_to_array
from smac.epm.rf_with_instances import RandomForestWithInstances
from smac.optimizer.acquisition import AbstractAcquisitionFunction
from smac.optimizer.ei_optimization import AcquisitionFunctionMaximizer, RandomSearch
from smac.optimizer.random_configuration_chooser import (
    ChooserNoCoolDown,
    RandomConfigurationChooser,
)
from smac.runhistory.runhistory import RunHistory
from smac.runhistory.runhistory2epm import AbstractRunHistory2EPM
from smac.scenario.scenario import Scenario
from smac.stats.stats import Stats

__copyright__ = "Copyright 2021, AutoML.org Freiburg-Hannover"
__license__ = "3-clause BSD"


class EPMChooser(object):
    """Interface to train the EPM and generate next configurations.

    Parameters
    ----------
    scenario: smac.scenario.scenario.Scenario
        Scenario object
    stats: smac.stats.stats.Stats
        statistics object with configuration budgets
    runhistory: smac.runhistory.runhistory.RunHistory
        runhistory with all runs so far
    model: smac.epm.rf_with_instances.RandomForestWithInstances
        empirical performance model (right now, we support only
        RandomForestWithInstances)
    acq_optimizer: smac.optimizer.ei_optimization.AcquisitionFunctionMaximizer
        Optimizer of acquisition function.
    restore_incumbent: Configuration
        incumbent to be used from the start. ONLY used to restore states.
    rng: np.random.RandomState
        Random number generator
    random_configuration_chooser:
        Chooser for random configuration -- one of

        * ChooserNoCoolDown(modulus)
        * ChooserLinearCoolDown(start_modulus, modulus_increment, end_modulus)
    predict_x_best: bool
        Choose x_best for computing the acquisition function via the model instead of via the observations.
    min_samples_model: int
        Minimum number of samples to build a model
    """

    def __init__(
        self,
        scenario: Scenario,
        stats: Stats,
        runhistory: RunHistory,
        runhistory2epm: AbstractRunHistory2EPM,
        model: RandomForestWithInstances,
        acq_optimizer: AcquisitionFunctionMaximizer,
        acquisition_func: AbstractAcquisitionFunction,
        rng: np.random.RandomState,
        restore_incumbent: Configuration = None,
        random_configuration_chooser: RandomConfigurationChooser = ChooserNoCoolDown(modulus=2.0),
        predict_x_best: bool = True,
        min_samples_model: int = 1,
    ):
        self.logger = logging.getLogger(self.__module__ + "." + self.__class__.__name__)
        self.incumbent = restore_incumbent

        self.scenario = scenario
        self.stats = stats
        self.runhistory = runhistory
        self.rh2EPM = runhistory2epm
        self.model = model
        self.acq_optimizer = acq_optimizer
        self.acquisition_func = acquisition_func
        self.rng = rng
        self.random_configuration_chooser = random_configuration_chooser

        self._random_search = RandomSearch(
            acquisition_func,
            self.scenario.cs,  # type: ignore[attr-defined] # noqa F821
            rng,
        )

        self.initial_design_configs = []  # type: List[Configuration]

        self.predict_x_best = predict_x_best

        self.min_samples_model = min_samples_model
        self.currently_considered_budgets = [
            0.0,
        ]

    def _collect_data_to_train_model(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        # if we use a float value as a budget, we want to train the model only on the highest budget
        available_budgets = []
        for run_key in self.runhistory.data.keys():
            available_budgets.append(run_key.budget)

        # Sort available budgets from highest to lowest budget
        available_budgets = sorted(list(set(available_budgets)), reverse=True)

        # Get #points per budget and if there are enough samples, then build a model
        for b in available_budgets:
            X, Y = self.rh2EPM.transform(
                self.runhistory,
                budget_subset=[
                    b,
                ],
            )
            if X.shape[0] >= self.min_samples_model:
                self.currently_considered_budgets = [
                    b,
                ]
                configs_array = self.rh2EPM.get_configurations(
                    self.runhistory, budget_subset=self.currently_considered_budgets
                )
                return X, Y, configs_array

        return (
            np.empty(shape=[0, 0]),
            np.empty(
                shape=[
                    0,
                ]
            ),
            np.empty(shape=[0, 0]),
        )

    def _get_evaluated_configs(self) -> List[Configuration]:
        return self.runhistory.get_all_configs_per_budget(budget_subset=self.currently_considered_budgets)

    def choose_next(self, incumbent_value: float = None) -> Iterator[Configuration]:
        """Choose next candidate solution with Bayesian optimization. The suggested configurations
        depend on the argument ``acq_optimizer`` to the ``SMBO`` class.

        Parameters
        ----------
        incumbent_value: float
            Cost value of incumbent configuration (required for acquisition function);
            If not given, it will be inferred from runhistory or predicted;
            if not given and runhistory is empty, it will raise a ValueError.

        Returns
        -------
        Iterator
        """
        self.logger.debug("Search for next configuration")
        X, Y, X_configurations = self._collect_data_to_train_model()

        if X.shape[0] == 0:
            # Only return a single point to avoid an overly high number of
            # random search iterations
            return self._random_search.maximize(runhistory=self.runhistory, stats=self.stats, num_points=1)
        self.model.train(X, Y)

        if incumbent_value is not None:
            best_observation = incumbent_value
            x_best_array = None  # type: Optional[np.ndarray]
        else:
            if self.runhistory.empty():
                raise ValueError("Runhistory is empty and the cost value of " "the incumbent is unknown.")
            x_best_array, best_observation = self._get_x_best(self.predict_x_best, X_configurations)

        self.acquisition_func.update(
            model=self.model,
            eta=best_observation,
            incumbent_array=x_best_array,
            num_data=len(self._get_evaluated_configs()),
            X=X_configurations,
        )

        challengers = self.acq_optimizer.maximize(
            runhistory=self.runhistory,
            stats=self.stats,
            num_points=self.scenario.acq_opt_challengers,  # type: ignore[attr-defined] # noqa F821
            random_configuration_chooser=self.random_configuration_chooser,
        )
        return challengers

    def _get_x_best(self, predict: bool, X: np.ndarray) -> Tuple[np.ndarray, float]:
        """Get value, configuration, and array representation of the "best" configuration.

        The definition of best varies depending on the argument ``predict``. If set to ``True``,
        this function will return the stats of the best configuration as predicted by the model,
        otherwise it will return the stats for the best observed configuration.

        Parameters
        ----------
        predict : bool
            Whether to use the predicted or observed best.

        Returns
        -------
        float
        np.ndarry
        Configuration
        """
        if predict:
            costs = list(
                map(
                    lambda x: (
                        self.model.predict_marginalized_over_instances(x.reshape((1, -1)))[0][0][0],
                        x,
                    ),
                    X,
                )
            )
            costs = sorted(costs, key=lambda t: t[0])
            x_best_array = costs[0][1]
            best_observation = costs[0][0]
            # won't need log(y) if EPM was already trained on log(y)
        else:
            all_configs = self.runhistory.get_all_configs_per_budget(budget_subset=self.currently_considered_budgets)
            x_best = self.incumbent
            x_best_array = convert_configurations_to_array(all_configs)
            best_observation = self.runhistory.get_cost(x_best)
            best_observation_as_array = np.array(best_observation).reshape((1, 1))
            # It's unclear how to do this for inv scaling and potential future scaling.
            # This line should be changed if necessary
            best_observation = self.rh2EPM.transform_response_values(best_observation_as_array)
            best_observation = best_observation[0][0]

        return x_best_array, best_observation



from typing import Callable, Dict, List, Optional, Type, Union

import logging

import dask.distributed  # type: ignore
import numpy as np

from smac.configspace import Configuration
from smac.epm.random_epm import RandomEPM
from smac.facade.smac_ac_facade import SMAC4AC
from smac.initial_design.initial_design import InitialDesign
from smac.intensification.abstract_racer import AbstractRacer
from smac.optimizer.ei_optimization import AcquisitionFunctionMaximizer, RandomSearch
from smac.optimizer.multi_objective.abstract_multi_objective_algorithm import (
    AbstractMultiObjectiveAlgorithm,
)
from smac.runhistory.runhistory import RunHistory
from smac.runhistory.runhistory2epm import (
    AbstractRunHistory2EPM,
    RunHistory2EPM4Cost,
    RunHistory2EPM4LogCost,
)
from smac.scenario.scenario import Scenario
from smac.stats.stats import Stats
from smac.tae.base import BaseRunner

__author__ = "Marius Lindauer"
__copyright__ = "Copyright 2016, ML4AAD"
__license__ = "3-clause BSD"


class ROAR(SMAC4AC):
    """Facade to use ROAR mode.

    Parameters
    ----------
    scenario: smac.scenario.scenario.Scenario
        Scenario object
    tae_runner: smac.tae.base.BaseRunner or callable
        Callable or implementation of
        :class:`~smac.tae.base.BaseRunner`. In case a
        callable is passed it will be wrapped by
        :class:`~smac.tae.execute_func.ExecuteTAFuncDict`.
        If not set, it will be initialized with the
        :class:`~smac.tae.execute_ta_run_old.ExecuteTARunOld`.
    tae_runner_kwargs: Optional[Dict]
        arguments passed to constructor of '~tae_runner'
    runhistory: RunHistory
        Runhistory to store all algorithm runs
    intensifier: AbstractRacer
        intensification object to issue a racing to decide the current incumbent
    intensifier_kwargs: Optional[Dict]
        arguments passed to the constructor of '~intensifier'
    acquisition_function_optimizer : ~smac.optimizer.ei_optimization.AcquisitionFunctionMaximizer
        Object that implements the :class:`~smac.optimizer.ei_optimization.AcquisitionFunctionMaximizer`.
        Will use :class:`smac.optimizer.ei_optimization.RandomSearch` if not set. Can be used
        to perform random search over a fixed set of configurations.
    acquisition_function_optimizer_kwargs: Optional[dict]
        Arguments passed to constructor of `~acquisition_function_optimizer`
    multi_objective_algorithm: Optional[Type["AbstractMultiObjectiveAlgorithm"]]
        Class that implements multi objective logic. If None, will use:
    smac.optimizer.multi_objective.aggregation_strategy.MeanAggregationStrategy
        Multi objective only becomes active if the objective
        specified in `~scenario.run_obj` is a List[str] with at least two entries.
    multi_objective_kwargs: Optional[Dict]
        Arguments passed to `~multi_objective_algorithm`.
    initial_design : InitialDesign
        initial sampling design
    initial_design_kwargs: Optional[dict]
        arguments passed to constructor of `~initial_design`
    initial_configurations: List[Configuration]
        list of initial configurations for initial design --
        cannot be used together with initial_design
    stats: Stats
        optional stats object
    rng: Optional[int, np.random.RandomState]
        Random number generator
    run_id: Optional[int]
        Run ID will be used as subfolder for output_dir.
    dask_client : dask.distributed.Client
        User-created dask client, can be used to start a dask cluster and then attach SMAC to it.
    n_jobs : int, optional
        Number of jobs. If > 1 or -1, this creates a dask client if ``dask_client`` is ``None``. Will
        be ignored if ``dask_client`` is not ``None``.
        If ``None``, this value will be set to 1, if ``-1``, this will be set to the number of cpu cores.

    Attributes
    ----------
    logger

    See Also
    --------
    :class:`~smac.facade.smac_ac_facade.SMAC4AC`
    """

    def __init__(
        self,
        scenario: Scenario,
        tae_runner: Optional[Union[Type[BaseRunner], Callable]] = None,
        tae_runner_kwargs: Optional[Dict] = None,
        runhistory: RunHistory = None,
        intensifier: Optional[Type[AbstractRacer]] = None,
        intensifier_kwargs: Optional[Dict] = None,
        acquisition_function_optimizer: Optional[Type[AcquisitionFunctionMaximizer]] = None,
        acquisition_function_optimizer_kwargs: Optional[dict] = None,
        multi_objective_algorithm: Optional[Type[AbstractMultiObjectiveAlgorithm]] = None,
        multi_objective_kwargs: Optional[Dict] = None,
        initial_design: Optional[Type[InitialDesign]] = None,
        initial_design_kwargs: Optional[dict] = None,
        initial_configurations: List[Configuration] = None,
        stats: Stats = None,
        rng: Optional[Union[int, np.random.RandomState]] = None,
        run_id: Optional[int] = None,
        dask_client: Optional[dask.distributed.Client] = None,
        n_jobs: Optional[int] = 1,
    ):
        self.logger = logging.getLogger(self.__module__ + "." + self.__class__.__name__)

        scenario.acq_opt_challengers = 1  # type: ignore[attr-defined] # noqa F821

        if acquisition_function_optimizer is None:
            acquisition_function_optimizer = RandomSearch

        if scenario.run_obj == "runtime":
            # We need to do this to be on the same scale for imputation (although we only impute with a Random EPM)
            runhistory2epm = RunHistory2EPM4LogCost  # type: Type[AbstractRunHistory2EPM]
        else:
            runhistory2epm = RunHistory2EPM4Cost

        # use SMAC facade
        super().__init__(
            scenario=scenario,
            tae_runner=tae_runner,
            tae_runner_kwargs=tae_runner_kwargs,
            runhistory=runhistory,
            intensifier=intensifier,
            intensifier_kwargs=intensifier_kwargs,
            runhistory2epm=runhistory2epm,
            multi_objective_algorithm=multi_objective_algorithm,
            multi_objective_kwargs=multi_objective_kwargs,
            initial_design=initial_design,
            initial_design_kwargs=initial_design_kwargs,
            initial_configurations=initial_configurations,
            run_id=run_id,
            acquisition_function_optimizer=acquisition_function_optimizer,
            acquisition_function_optimizer_kwargs=acquisition_function_optimizer_kwargs,
            model=RandomEPM,
            rng=rng,
            stats=stats,
            dask_client=dask_client,
            n_jobs=n_jobs,
        )



from ConfigSpace.configuration_space import ConfigurationSpace
from ConfigSpace.hyperparameters import UniformFloatHyperparameter

__copyright__ = "Copyright 2021, AutoML.org Freiburg-Hannover"
__license__ = "3-clause BSD"


def get_branin_config_space() -> ConfigurationSpace:
    """Returns the branin configspace."""
    cs = ConfigurationSpace()
    cs.add_hyperparameter(UniformFloatHyperparameter("x", -5, 10))
    cs.add_hyperparameter(UniformFloatHyperparameter("y", 0, 15))
    return cs



from typing import Any, Dict, List, Tuple

from smac.configspace import ConfigurationSpace
from smac.epm.base_epm import AbstractEPM
from smac.epm.base_uncorrelated_mo_model import UncorrelatedMultiObjectiveModel
from smac.epm.rf_with_instances import RandomForestWithInstances

__copyright__ = "Copyright 2021, AutoML.org Freiburg-Hannover"
__license__ = "3-clause BSD"


class UncorrelatedMultiObjectiveRandomForestWithInstances(UncorrelatedMultiObjectiveModel):
    """Wrapper for the random forest to predict multiple targets.

    Only a list with the target names and the types array for the
    underlying forest model are mandatory. All other hyperparameters to
    the random forest can be passed via kwargs. Consult the documentation of
    the random forest for the hyperparameters and their meanings.
    """

    def construct_estimators(
        self,
        configspace: ConfigurationSpace,
        types: List[int],
        bounds: List[Tuple[float, float]],
        model_kwargs: Dict[str, Any],
    ) -> List[AbstractEPM]:
        """
        Construct a list of estimators. The number of the estimators equals 'self.num_targets'
        Parameters
        ----------
        configspace : ConfigurationSpace
            Configuration space to tune for.
        types : List[int]
            Specifies the number of categorical values of an input dimension where
            the i-th entry corresponds to the i-th input dimension. Let's say we
            have 2 dimension where the first dimension consists of 3 different
            categorical choices and the second dimension is continuous than we
            have to pass [3, 0]. Note that we count starting from 0.
        bounds : List[Tuple[float, float]]
            bounds of input dimensions: (lower, uppper) for continuous dims; (n_cat, np.nan) for categorical dims
        model_kwargs : Dict[str, Any]
            model kwargs for initializing models
        Returns
        -------
        estimators: List[AbstractEPM]
            A list of Random Forests
        """
        return [RandomForestWithInstances(configspace, types, bounds, **model_kwargs) for _ in range(self.num_targets)]



from abc import ABC
from typing import Optional

import numpy as np


class AbstractMultiObjectiveAlgorithm(ABC):
    """
    A general interface for multi-objective optimizer, depending on different strategies.
    It can be applied to rh2epm or epmchooser.
    """

    def __init__(self, num_obj: int, rng: Optional[np.random.RandomState] = None):

        if rng is None:
            rng = np.random.RandomState(0)

        self.num_obj = num_obj
        self.rng = rng



from typing import Any

from smac.facade.roar_facade import ROAR
from smac.initial_design.random_configuration_design import RandomConfigurations
from smac.intensification.hyperband import Hyperband

__author__ = "Ashwin Raaghav Narayanan"
__copyright__ = "Copyright 2019, ML4AAD"
__license__ = "3-clause BSD"


class HB4AC(ROAR):
    """Facade to use model-free Hyperband for algorithm configuration.

    This facade overwrites options available via the SMAC facade.

    See Also
    --------
    :class:`~smac.facade.smac_ac_facade.SMAC4AC` for documentation of parameters.

    Attributes
    ----------
    logger
    stats : Stats
    solver : SMBO
    runhistory : RunHistory
        List with information about previous runs
    trajectory : list
        List of all incumbents
    """

    def __init__(self, **kwargs: Any):
        kwargs["initial_design"] = kwargs.get("initial_design", RandomConfigurations)

        # Intensification parameters
        # select Hyperband as the intensifier ensure respective parameters are provided
        kwargs["intensifier"] = Hyperband

        # set Hyperband parameters if not given
        intensifier_kwargs = kwargs.get("intensifier_kwargs", dict())
        intensifier_kwargs["min_chall"] = 1
        if intensifier_kwargs.get("eta") is None:
            intensifier_kwargs["eta"] = 3
        if intensifier_kwargs.get("instance_order") is None:
            intensifier_kwargs["instance_order"] = "shuffle_once"
        kwargs["intensifier_kwargs"] = intensifier_kwargs

        super().__init__(**kwargs)
        self.logger.info(self.__class__)



import os
import shutil
from logging import Logger

from smac.scenario.scenario import Scenario

__copyright__ = "Copyright 2021, AutoML.org Freiburg-Hannover"
__license__ = "3-clause BSD"


def create_output_directory(
    scenario: Scenario,
    run_id: int,
    logger: Logger = None,
) -> str:
    """Create output directory for this run.

    Side effect: Adds the current output directory to the scenario object!

    Parameters
    ----------
    scenario : ~smac.scenario.scenario.Scenario
    run_id : int

    Returns
    -------
    str
    """
    if scenario.output_dir:  # type: ignore[attr-defined] # noqa F821
        output_dir = os.path.join(
            scenario.output_dir,  # type: ignore[attr-defined] # noqa F821
            "run_%d" % (run_id),
        )
    else:
        return ""
    if os.path.exists(output_dir):
        move_to = output_dir + ".OLD"
        while os.path.exists(move_to):
            move_to += ".OLD"
        shutil.move(output_dir, move_to)
        if logger is not None:
            logger.warning(
                'Output directory "%s" already exists! ' 'Moving old folder to "%s".',
                output_dir,
                move_to,
            )
    scenario.output_dir_for_this_run = output_dir
    return output_dir



from typing import List

import warnings

from ConfigSpace.configuration_space import Configuration
from ConfigSpace.hyperparameters import Constant
from scipy.stats.qmc import Sobol

from smac.initial_design.initial_design import InitialDesign

__author__ = "Marius Lindauer"
__copyright__ = "Copyright 2018, ML4AAD"
__license__ = "3-clause BSD"


class SobolDesign(InitialDesign):
    """Sobol sequence design with a scrambled Sobol sequence.

    See https://scipy.github.io/devdocs/reference/generated/scipy.stats.qmc.Sobol.html for further information

    Attributes
    ----------
    configs : List[Configuration]
        List of configurations to be evaluated
        Don't pass configs to the constructor;
        otherwise factorial design is overwritten
    """

    def _select_configurations(self) -> List[Configuration]:
        """Selects a single configuration to run.

        Returns
        -------
        config: Configuration
            initial incumbent configuration
        """
        params = self.cs.get_hyperparameters()

        constants = 0
        for p in params:
            if isinstance(p, Constant):
                constants += 1

        dim = len(params) - constants
        sobol_gen = Sobol(d=dim, scramble=True, seed=self.rng.randint(low=0, high=10000000))

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            sobol = sobol_gen.random(self.init_budget)

        return self._transform_continuous_designs(design=sobol, origin="Sobol", cs=self.cs)



from typing import Any, Dict, List, Optional, Sequence, Tuple, Union, cast

import logging
import os
from collections import namedtuple

import numpy as np
from joblib import Parallel, delayed

from smac.configspace import Configuration, convert_configurations_to_array
from smac.epm.rf_with_instances import RandomForestWithInstances
from smac.epm.rfr_imputator import RFRImputator
from smac.epm.util_funcs import get_types
from smac.runhistory.runhistory import RunHistory, RunInfo, RunKey, RunValue, StatusType
from smac.runhistory.runhistory2epm import RunHistory2EPM4Cost
from smac.scenario.scenario import Scenario
from smac.stats.stats import Stats
from smac.tae.base import BaseRunner
from smac.tae.execute_ta_run_old import ExecuteTARunOld
from smac.utils.constants import MAXINT

__author__ = "Joshua Marben"
__copyright__ = "Copyright 2017, ML4AAD"
__license__ = "3-clause BSD"
__maintainer__ = "Joshua Marben"
__email__ = "joshua.marben@neptun.uni-freiburg.de"


def _unbound_tae_starter(
    tae: BaseRunner,
    runhistory: Optional[RunHistory],
    run_info: RunInfo,
    *args: Any,
    **kwargs: Any,
) -> RunValue:
    """Unbound function to be used by joblibs Parallel, since directly passing the TAE results in
    pickling-problems.

    Parameters
    ----------
    tae: BaseRunner
        tae to be used
    runhistory: RunHistory
        runhistory to save
    run_info: RunInfo
        Config to be launched
    *args, **kwargs: various
        arguments to the tae

    Returns
    -------
    tae_results: RunValue
        return from tae.start
    """
    run_info, result = tae.run_wrapper(run_info)
    tae.stats.submitted_ta_runs += 1
    tae.stats.finished_ta_runs += 1
    tae.stats.ta_time_used += float(result.time)
    if runhistory:
        runhistory.add(
            config=run_info.config,
            cost=result.cost,
            time=result.time,
            status=result.status,
            instance_id=run_info.instance,
            seed=run_info.seed,
            budget=run_info.budget,
        )
        tae.stats.n_configs = len(runhistory.config_ids)

    return result


_Run = namedtuple("_Run", "config inst seed inst_specs")


class Validator(object):
    """Validator for the output of SMAC-scenarios.

    Evaluates specified configurations on specified instances.

    Parameters
    ----------
    scenario: Scenario
        scenario object for cutoff, instances, features and specifics
    trajectory: trajectory-list
        trajectory to take incumbent(s) from
    rng: np.random.RandomState or int
        Random number generator or seed
    """

    def __init__(
        self,
        scenario: Scenario,
        trajectory: Optional[List],
        rng: Union[np.random.RandomState, int, None] = None,
    ) -> None:
        self.logger = logging.getLogger(self.__module__ + "." + self.__class__.__name__)

        self.traj = trajectory
        self.scen = scenario
        self.epm = None  # type: Optional[RandomForestWithInstances]

        if isinstance(rng, np.random.RandomState):
            self.rng = rng
        elif isinstance(rng, int):
            self.rng = np.random.RandomState(seed=rng)
        else:
            self.logger.debug("no seed given, using default seed of 1")
            num_run = 1
            self.rng = np.random.RandomState(seed=num_run)

    def _save_results(
        self,
        rh: RunHistory,
        output_fn: Optional[str],
        backup_fn: Optional[str] = None,
    ) -> None:
        """Helper to save results to file.

        Parameters
        ----------
        rh: RunHistory
            runhistory to save
        output_fn: str
            if ends on '.json': filename to save history to
            else: directory to save runhistory to (filename is backup_fn)
        backup_fn: str
            if output_fn does not end on '.json', treat output_fn as dir and
            append backup_fn as filename (if output_fn ends on '.json', this
            argument is ignored)
        """
        if not output_fn:
            self.logger.info("No output specified, validated runhistory not saved.")
            return
        # Check if a folder or a file is specified as output
        if not output_fn.endswith(".json"):
            if backup_fn is None:
                raise ValueError("If output_fn does not end with .json the argument backup_fn needs to be given.")
            output_dir = output_fn
            output_fn = os.path.join(output_dir, backup_fn)
            self.logger.debug('Output is "%s", changing to "%s"!', output_dir, output_fn)
        base = os.path.split(output_fn)[0]
        if not base == "" and not os.path.exists(base):
            self.logger.debug('Folder ("%s") doesn\'t exist, creating.', base)
            os.makedirs(base)
        rh.save_json(output_fn)
        self.logger.info("Saving validation-results in %s", output_fn)

    def validate(
        self,
        config_mode: Union[str, List[Configuration]] = "def",
        instance_mode: Union[str, List[str]] = "test",
        repetitions: int = 1,
        n_jobs: int = 1,
        backend: str = "threading",
        runhistory: Optional[RunHistory] = None,
        tae: BaseRunner = None,
        output_fn: Optional[str] = None,
    ) -> RunHistory:
        """Validate configs on instances and save result in runhistory. If a runhistory is provided
        as input it is important that you run it on the same/comparable hardware.

        side effect: if output is specified, saves runhistory to specified
        output directory.

        Parameters
        ----------
        config_mode: str or list<Configuration>
            string or directly a list of Configuration.
            string from [def, inc, def+inc, wallclock_time, cpu_time, all].
            time evaluates at cpu- or wallclock-timesteps of:
            [max_time/2^0, max_time/2^1, max_time/2^3, ..., default]
            with max_time being the highest recorded time
        instance_mode: str or list<str>
            what instances to use for validation, either from
            [train, test, train+test] or directly a list of instances
        repetitions: int
            number of repetitions in nondeterministic algorithms
        n_jobs: int
            number of parallel processes used by joblib
        backend: str
            what backend joblib should use for parallel runs
        runhistory: RunHistory
            optional, RunHistory-object to reuse runs
        tae: BaseRunner
            tae to be used. if None, will initialize ExecuteTARunOld
        output_fn: str
            path to runhistory to be saved. if the suffix is not '.json', will
            be interpreted as directory and filename will be
            'validated_runhistory.json'

        Returns
        -------
        runhistory: RunHistory
            runhistory with validated runs
        """
        self.logger.debug(
            "Validating configs '%s' on instances '%s', repeating %d times" " with %d parallel runs on backend '%s'.",
            config_mode,
            instance_mode,
            repetitions,
            n_jobs,
            backend,
        )

        # Get all runs to be evaluated as list
        runs, validated_rh = self._get_runs(config_mode, instance_mode, repetitions, runhistory)

        # Create new Stats without limits
        inf_scen = Scenario(
            {
                "run_obj": self.scen.run_obj,
                "cutoff_time": self.scen.cutoff,  # type: ignore[attr-defined] # noqa F821
                "output_dir": "",
            }
        )
        inf_stats = Stats(inf_scen)
        inf_stats.start_timing()

        # Create TAE
        if not tae:
            tae = ExecuteTARunOld(
                ta=self.scen.ta,  # type: ignore[attr-defined] # noqa F821
                stats=inf_stats,
                run_obj=self.scen.run_obj,
                par_factor=self.scen.par_factor,  # type: ignore[attr-defined] # noqa F821
                cost_for_crash=self.scen.cost_for_crash,
            )  # type: ignore[attr-defined] # noqa F821
        else:
            # Inject endless-stats
            tae.stats = inf_stats

        # Validate!
        run_results = self._validate_parallel(tae, runs, n_jobs, backend, runhistory)
        assert len(run_results) == len(runs), (run_results, runs)

        # tae returns (status, cost, runtime, additional_info)
        # Add runs to RunHistory
        for run, result in zip(runs, run_results):
            validated_rh.add(
                config=run.config,
                cost=result.cost,
                time=result.time,
                status=result.status,
                instance_id=run.inst,
                seed=run.seed,
                additional_info=result.additional_info,
            )

        self._save_results(validated_rh, output_fn, backup_fn="validated_runhistory.json")
        return validated_rh

    def _validate_parallel(
        self,
        tae: BaseRunner,
        runs: List[_Run],
        n_jobs: int,
        backend: str,
        runhistory: Optional[RunHistory] = None,
    ) -> List[RunValue]:
        """Validate runs with joblibs Parallel-interface.

        Parameters
        ----------
        tae: BaseRunner
            tae to be used for validation
        runs: list<_Run>
            list with _Run-objects
            [_Run(config=CONFIG1,inst=INSTANCE1,seed=SEED1,inst_specs=INST_SPECIFICS1), ...]
        n_jobs: int
            number of cpus to use for validation (-1 to use all)
        backend: str
            what backend to use for parallelization
        runhistory: RunHistory
            optional, RunHistory-object to reuse runs

        Returns
        -------
        run_results: list<tuple(tae-returns)>
            results as returned by tae
        """
        # Runs with parallel
        run_results = Parallel(n_jobs=n_jobs, backend=backend)(
            delayed(_unbound_tae_starter)(
                tae,
                runhistory,
                RunInfo(
                    config=run.config,
                    instance=run.inst,
                    instance_specific="0",
                    seed=run.seed,
                    cutoff=self.scen.cutoff,  # type: ignore[attr-defined] # noqa F821
                    capped=False,
                    budget=0,
                ),
            )
            for run in runs
        )
        return run_results

    def validate_epm(
        self,
        config_mode: Union[str, List[Configuration]] = "def",
        instance_mode: Union[str, List[str]] = "test",
        repetitions: int = 1,
        runhistory: Optional[RunHistory] = None,
        output_fn: Optional[str] = None,
        reuse_epm: bool = True,
    ) -> RunHistory:
        """Use EPM to predict costs/runtimes for unknown config/inst-pairs.

        side effect: if output is specified, saves runhistory to specified
        output directory.

        Parameters
        ----------
        output_fn: str
            path to runhistory to be saved. if the suffix is not '.json', will
            be interpreted as directory and filename will be
            'validated_runhistory_EPM.json'
        config_mode: str or list<Configuration>
            string or directly a list of Configuration, string from [def, inc, def+inc, wallclock_time, cpu_time, all].
            time evaluates at cpu- or wallclock-timesteps of:
            [max_time/2^0, max_time/2^1, max_time/2^3, ..., default] with max_time being the highest recorded time
        instance_mode: str or list<str>
            what instances to use for validation, either from
            [train, test, train+test] or directly a list of instances
        repetitions: int
            number of repetitions in nondeterministic algorithms
        runhistory: RunHistory
            optional, RunHistory-object to reuse runs
        reuse_epm: bool
            if true (and if `self.epm`), reuse epm to validate runs

        Returns
        -------
        runhistory: RunHistory
            runhistory with predicted runs
        """
        if not isinstance(runhistory, RunHistory) and (self.epm is None or not reuse_epm):
            raise ValueError("No runhistory specified for validating with EPM!")
        elif not reuse_epm or self.epm is None:
            # Create RandomForest
            types, bounds = get_types(self.scen.cs, self.scen.feature_array)  # type: ignore[attr-defined] # noqa F821
            epm = RandomForestWithInstances(
                configspace=self.scen.cs,  # type: ignore[attr-defined] # noqa F821
                types=types,
                bounds=bounds,
                instance_features=self.scen.feature_array,
                seed=self.rng.randint(MAXINT),
                ratio_features=1.0,
            )
            # Use imputor if objective is runtime
            imputor = None
            impute_state = None
            impute_censored_data = False
            if self.scen.run_obj == "runtime":
                threshold = self.scen.cutoff * self.scen.par_factor  # type: ignore[attr-defined] # noqa F821
                imputor = RFRImputator(
                    rng=self.rng,
                    cutoff=self.scen.cutoff,  # type: ignore[attr-defined] # noqa F821
                    threshold=threshold,
                    model=epm,
                )
                impute_censored_data = True
                impute_state = [StatusType.CAPPED]
                success_states = [
                    StatusType.SUCCESS,
                ]
            else:
                success_states = [StatusType.SUCCESS, StatusType.CRASHED, StatusType.MEMOUT]

            # Transform training data (from given rh)
            rh2epm = RunHistory2EPM4Cost(
                num_params=len(self.scen.cs.get_hyperparameters()),  # type: ignore[attr-defined] # noqa F821
                scenario=self.scen,
                rng=self.rng,
                impute_censored_data=impute_censored_data,
                imputor=imputor,
                impute_state=impute_state,
                success_states=success_states,
            )
            assert runhistory is not None  # please mypy
            X, y = rh2epm.transform(runhistory)
            self.logger.debug("Training model with data of shape X: %s, y:%s", str(X.shape), str(y.shape))
            # Train random forest
            epm.train(X, y)
        else:
            epm = cast(RandomForestWithInstances, self.epm)

        # Predict desired runs
        runs, rh_epm = self._get_runs(config_mode, instance_mode, repetitions, runhistory)

        feature_array_size = len(self.scen.cs.get_hyperparameters())  # type: ignore[attr-defined] # noqa F821
        if self.scen.feature_array is not None:
            feature_array_size += self.scen.feature_array.shape[1]

        X_pred = np.empty((len(runs), feature_array_size))
        for idx, run in enumerate(runs):
            if self.scen.feature_array is not None and run.inst is not None:
                X_pred[idx] = np.hstack(
                    [
                        convert_configurations_to_array([run.config])[0],
                        self.scen.feature_dict[run.inst],  # type: ignore
                    ]
                )
            else:
                X_pred[idx] = convert_configurations_to_array([run.config])[0]
        self.logger.debug("Predicting desired %d runs, data has shape %s", len(runs), str(X_pred.shape))

        y_pred = epm.predict(X_pred)
        self.epm = epm

        # Add runs to runhistory
        for run, pred in zip(runs, y_pred[0]):
            rh_epm.add(
                config=run.config,
                cost=float(pred),
                time=float(pred),
                status=StatusType.SUCCESS,
                instance_id=run.inst,
                seed=-1,
                additional_info={"additional_info": "ESTIMATED USING EPM!"},
            )

        if output_fn:
            self._save_results(rh_epm, output_fn, backup_fn="validated_runhistory_EPM.json")
        return rh_epm

    def _get_runs(
        self,
        configs: Union[str, List[Configuration]],
        insts: Union[str, List[str]],
        repetitions: int = 1,
        runhistory: RunHistory = None,
    ) -> Tuple[List[_Run], RunHistory]:
        """Generate list of SMAC-TAE runs to be executed. This means combinations of configs with
        all instances on a certain number of seeds.

        side effect: Adds runs that don't need to be reevaluated to self.rh!

        Parameters
        ----------
        configs: str or list<Configuration>
            string or directly a list of Configuration
            str from [def, inc, def+inc, wallclock_time, cpu_time, all]
            time evaluates at cpu- or wallclock-timesteps of:
            [max_time/2^0, max_time/2^1, max_time/2^3, ..., default]
            with max_time being the highest recorded time
        insts: str or list<str>
            what instances to use for validation, either from
            [train, test, train+test] or directly a list of instances
        repetitions: int
            number of seeds per instance/config-pair to be evaluated
        runhistory: RunHistory
            optional, try to reuse this runhistory and save some runs

        Returns
        -------
        runs: list<_Run>
            list with _Runs
            [_Run(config=CONFIG1,inst=INSTANCE1,seed=SEED1,inst_specs=INST_SPECIFICS1),
             _Run(config=CONFIG2,inst=INSTANCE2,seed=SEED2,inst_specs=INST_SPECIFICS2),
             ...]
        """
        # Get relevant configurations and instances
        if isinstance(configs, str):
            configs = self._get_configs(configs)
        if isinstance(insts, str):
            instances = sorted(self._get_instances(insts))  # type: Sequence[Union[str, None]]
        elif insts is not None:
            instances = sorted(insts)
        else:
            instances = [None]
        # If no instances are given, fix the instances to one "None" instance
        if not instances:
            instances = [None]

        # If algorithm is deterministic, fix repetitions to 1
        if self.scen.deterministic and repetitions != 1:  # type: ignore[attr-defined] # noqa F821
            self.logger.warning(
                "Specified %d repetitions, but fixing to 1, " "because algorithm is deterministic.",
                repetitions,
            )
            repetitions = 1

        # Extract relevant information from given runhistory
        inst_seed_config = self._process_runhistory(configs, instances, runhistory)

        # Now create the actual run-list
        runs = []
        # Counter for runs without the need of recalculation
        runs_from_rh = 0
        # If we reuse runs, we want to return them as well
        new_rh = RunHistory()

        for i in instances:
            for rep in range(repetitions):
                # First, find a seed and add all the data we can take from the
                # given runhistory to "our" validation runhistory.
                configs_evaluated = []  # type: Configuration
                if runhistory and i in inst_seed_config:
                    # Choose seed based on most often evaluated inst-seed-pair
                    seed, configs_evaluated = inst_seed_config[i].pop(0)
                    # Delete inst if all seeds are used
                    if not inst_seed_config[i]:
                        inst_seed_config.pop(i)
                    # Add runs to runhistory
                    for c in configs_evaluated[:]:
                        runkey = RunKey(runhistory.config_ids[c], i, seed)
                        cost, time, status, start, end, additional_info = runhistory.data[runkey]
                        if status in [StatusType.CRASHED, StatusType.ABORT, StatusType.CAPPED]:
                            # Not properly executed target algorithm runs should be repeated
                            configs_evaluated.remove(c)
                            continue
                        new_rh.add(
                            c,
                            cost,
                            time,
                            status,
                            instance_id=i,
                            seed=seed,
                            starttime=start,
                            endtime=end,
                            additional_info=additional_info,
                        )
                        runs_from_rh += 1
                else:
                    # If no runhistory or no entries for instance, get new seed
                    seed = self.rng.randint(MAXINT)

                # We now have a seed and add all configs that are not already
                # evaluated on that seed to the runs-list. This way, we
                # guarantee the same inst-seed-pairs for all configs.
                for config in [c for c in configs if c not in configs_evaluated]:
                    # Only use specifics if specific exists, else use string "0"
                    specs = self.scen.instance_specific[i] if i and i in self.scen.instance_specific else "0"
                    runs.append(_Run(config=config, inst=i, seed=seed, inst_specs=specs))

        self.logger.info(
            "Collected %d runs from %d configurations on %d "
            "instances with %d repetitions. Reusing %d runs from "
            "given runhistory.",
            len(runs),
            len(configs),
            len(instances),
            repetitions,
            runs_from_rh,
        )

        return runs, new_rh

    def _process_runhistory(
        self,
        configs: List[Configuration],
        insts: Sequence[Optional[str]],
        runhistory: Optional[RunHistory],
    ) -> Dict[str, List[Tuple[int, List[Configuration]]]]:
        """Processes runhistory from self._get_runs by extracting already evaluated (relevant)
        config-inst-seed tuples.

        Parameters
        ----------
        configs: list(Configuration)
            list of configs of interest
        insts: list(str)
            list of instances of interest
        runhistory: RunHistory
            runhistory to extract runs from

        Returns
        -------
        inst_seed_config: dict<str : list(tuple(int, tuple(configs)))>
            dictionary mapping instances to a list of tuples of already used
            seeds and the configs that this inst-seed-pair has been evaluated
            on, sorted by the number of configs
        """
        # We want to reuse seeds that have been used on most configurations
        # To this end, we create a dictionary as {instances:{seed:[configs]}}
        # Like this we can easily retrieve the most used instance-seed pairs to
        # minimize the number of runs to be evaluated
        if runhistory:
            inst_seed_config = {}  # type: Dict[str, Dict[int, List[Configuration]]]
            relevant = dict()
            for key in runhistory.data:
                if runhistory.ids_config[key.config_id] in configs and key.instance_id in insts:
                    relevant[key] = runhistory.data[key]

            # Change data-structure to {instances:[(seed1, (configs)), (seed2, (configs), ... ]}
            # to make most used seed easily accessible, we sort after length of configs
            for key in relevant:
                inst, seed = key.instance_id, key.seed
                config = runhistory.ids_config[key.config_id]
                if inst in inst_seed_config:
                    if seed in inst_seed_config[inst]:
                        inst_seed_config[inst][seed].append(config)
                    else:
                        inst_seed_config[inst][seed] = [config]
                else:
                    inst_seed_config[inst] = {seed: [config]}

            return {
                i: sorted(
                    [(seed, list(inst_seed_config[i][seed])) for seed in inst_seed_config[i]],
                    key=lambda x: len(x[1]),
                )
                for i in inst_seed_config
            }
        else:
            rval = {}  # type: Dict[str, List[Tuple[int, List[Configuration]]]]
            return rval

    def _get_configs(self, mode: str) -> List[str]:
        """Return desired configs.

        Parameters
        ----------
        mode: str
            str from [def, inc, def+inc, wallclock_time, cpu_time, all]
                time evaluates at cpu- or wallclock-timesteps of:
                [max_time/2^0, max_time/2^1, max_time/2^3, ..., default]
                with max_time being the highest recorded time

        Returns
        -------
        configs: list<Configuration>
            list with desired configurations
        """
        # Get trajectory and make sure it's not None to please mypy
        traj = self.traj
        assert traj is not None  # please mypy

        # Add desired configs
        configs = []
        mode = mode.lower()
        if mode not in ["def", "inc", "def+inc", "wallclock_time", "cpu_time", "all"]:
            raise ValueError("%s not a valid option for config_mode in validation." % mode)
        if mode == "def" or mode == "def+inc":
            configs.append(self.scen.cs.get_default_configuration())  # type: ignore[attr-defined] # noqa F821
        if mode == "inc" or mode == "def+inc":
            configs.append(traj[-1]["incumbent"])
        if mode in ["wallclock_time", "cpu_time"]:
            # get highest time-entry and add entries from there
            # not using wallclock_limit in case it's inf
            if mode == "wallclock_time" and np.isfinite(self.scen.wallclock_limit):
                max_time = self.scen.wallclock_limit
            elif mode == "cpu_time" and np.isfinite(self.scen.algo_runs_timelimit):
                max_time = self.scen.algo_runs_timelimit
            else:
                max_time = traj[-1][mode]
            counter = 2**0
            for entry in traj[::-1]:
                if entry[mode] <= max_time / counter and entry["incumbent"] not in configs:
                    configs.append(entry["incumbent"])
                    counter *= 2
            if not traj[0]["incumbent"] in configs:
                configs.append(traj[0]["incumbent"])  # add first
        if mode == "all":
            for entry in traj:
                if not entry["incumbent"] in configs:
                    configs.append(entry["incumbent"])
        self.logger.debug("Gathered %d configurations for mode %s.", len(configs), mode)
        return configs

    def _get_instances(self, mode: str) -> List[str]:
        """Get desired instances.

        Parameters
        ----------
        mode: str
            what instances to use for validation, from [train, test, train+test]

        Returns
        -------
        instances: list<str>
            instances to be used
        """
        instance_mode = mode.lower()
        if mode not in ["train", "test", "train+test"]:
            raise ValueError("%s not a valid option for instance_mode in validation." % mode)

        # Make sure if instances matter, than instances should be passed
        if (instance_mode == "train" and self.scen.train_insts == [None]) or (
            instance_mode == "test" and self.scen.test_insts == [None]
        ):
            self.logger.warning(
                "Instance mode is set to %s, but there are no "
                "%s-instances specified in the scenario. Setting instance mode to"
                '"train+test"!',
                instance_mode,
                instance_mode,
            )
            instance_mode = "train+test"

        instances = []  # type: List[str]
        if (instance_mode == "train" or instance_mode == "train+test") and not self.scen.train_insts == [None]:
            instances.extend(self.scen.train_insts)
        if (instance_mode == "test" or instance_mode == "train+test") and not self.scen.test_insts == [None]:
            instances.extend(self.scen.test_insts)
        return instances



from typing import List

from ConfigSpace import Configuration

from smac.initial_design.initial_design import InitialDesign

__author__ = "Marius Lindauer"
__copyright__ = "Copyright 2016, ML4AAD"
__license__ = "3-clause BSD"


class DefaultConfiguration(InitialDesign):
    """Initial design that evaluates default configuration."""

    def _select_configurations(self) -> List[Configuration]:
        """Selects the default configuration.

        Returns
        -------
        config: Configuration
            Initial incumbent configuration.
        """
        config = self.cs.get_default_configuration()
        config.origin = "Default"
        return [config]



from functools import partial

from ConfigSpace import (
    CategoricalHyperparameter,
    Configuration,
    ConfigurationSpace,
    Constant,
    InCondition,
    UniformFloatHyperparameter,
    UniformIntegerHyperparameter,
)
from ConfigSpace.exceptions import ForbiddenValueError
from ConfigSpace.read_and_write import json, pcs, pcs_new
from ConfigSpace.util import get_one_exchange_neighbourhood

from smac.configspace.util import convert_configurations_to_array

__copyright__ = "Copyright 2021, AutoML.org Freiburg-Hannover"
__license__ = "3-clause BSD"


__all__ = [
    "ConfigurationSpace",
    "Configuration",
    "Constant",
    "CategoricalHyperparameter",
    "UniformFloatHyperparameter",
    "UniformIntegerHyperparameter",
    "InCondition",
    "pcs",
    "pcs_new",
    "json",
    "get_one_exchange_neighbourhood",
    "convert_configurations_to_array",
    "ForbiddenValueError",
]

get_one_exchange_neighbourhood = partial(get_one_exchange_neighbourhood, stdev=0.05, num_neighbors=8)



from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from smac.optimizer.smbo import SMBO
from smac.runhistory.runhistory import RunInfo, RunValue

__copyright__ = "Copyright 2021, AutoML.org Freiburg-Hannover"
__license__ = "3-clause BSD"


class IncorporateRunResultCallback:
    """Callback to react on a new run result.

    Called after the finished run is added to the runhistory. Optionally
    return `False` to (gracefully) stop the optimization.
    """

    def __call__(
        self,
        smbo: "SMBO",
        run_info: RunInfo,
        result: RunValue,
        time_left: float,
    ) -> Optional[bool]:
        """Calls the callback."""
        ...



import os

import setuptools

from smac import (
    author,
    author_email,
    description,
    package_name,
    project_urls,
    url,
    version,
)

HERE = os.path.dirname(os.path.realpath(__file__))


def read_file(filepath: str) -> str:
    with open(filepath, "r", encoding="utf-8") as fh:
        return fh.read()


extras_require = {
    "dev": [
        "setuptools",
        "types-setuptools",
        # Test
        "pytest>=4.6",
        "pytest-cov",
        "pytest-xdist",
        "pytest-timeout",
        # Docs
        "automl-sphinx-theme>=0.1.9",
        # Others
        "mypy",
        "isort",
        "black",
        "pydocstyle",
        "flake8",
        "pre-commit",
    ],
}

setuptools.setup(
    name=package_name,
    author=author,
    author_email=author_email,
    description=description,
    long_description=read_file(os.path.join(HERE, "README.md")),
    long_description_content_type="text/markdown",
    license="BSD 3-Clause License",
    url=url,
    project_urls=project_urls,
    version=version,
    packages=setuptools.find_packages(exclude=["tests"]),
    include_package_data=True,
    python_requires=">=3.7",
    install_requires=[
        "numpy>=1.7.1",
        "scipy>=1.7.0",
        "psutil",
        "pynisher<1.0.0",
        "ConfigSpace>=0.5.0",
        "joblib",
        "scikit-learn>=0.22.0",
        "pyrfr>=0.8.0",
        "dask",
        "distributed",
        "emcee>=3.0.0",
    ],
    extras_require=extras_require,
    test_suite="pytest",
    platforms=["Linux"],
    entry_points={
        "console_scripts": ["smac = smac.smac_cli:cmd_line_call"],
    },
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Natural Language :: English",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
)
